#!"\xampp\perl\bin\perl.exe"
# server.pl (backend socketserver)
# (c) napp-it 2023,2024
use strict;
use vars qw(%data $timeout $rsize $comm $upsrv $max_connections $mem $sc $nocache $tm $line $ver $rl $cl $ml $nc $opt $haszfs $wpath $spath $cpath $on2 $r $command $allowed_cmd $allowed_ip $os $fn $on $dn $rh $hn $path $opt $filter $auth $tpath $wpath $dpath @q $ip $t $in %current $client);           # public varlist

###############################
$ver="24.09.03";
###############################

########################################################################
#################################### parameter #########################

# allowed client ip (single value or commalist)
# you can enter a complete ip like 192.168.1.2 or a substr like 192.168.1.
# if you enter a substr, all clients are allowed whose ip begins with this value
$allowed_ip="";

# Client timeout (read data from client, default 20)
$timeout=20;

# max client connections (read data from client, default 20)
$max_connections=20;

# callback method enables encrypted https transfers for commands and results (slower) instead unencrypted (faster)
#upsrv (Upload https Server for callback method)

# $data{'_upsrv'}= ";             # no callback or initiated by frontend web-gui
# $data{'_upsrv'}="ip";           # callback is enforced with calling ip (frontend Apache) as unvalidated https server (self signed certificate)
# $data{'_upsrv'}="hostname.dns"; # ex www2.universty.org, enforces a validated https server (public certificate)
$data{'_upsrv'}="";

# keyserver1 w1: webgui=default=calling ip)
$data{'_w1'}="";
$data{'_w1_auth'}="";                  # only if different to web-gui auth localhost~127.0.0.1

# keyserver2: w2
$data{'_w2'}="";
$data{'_w2_auth'}="";                  # only if different to web-gui auth localhost~127.0.0.1

##################################/ parameter ###########################
#########################################################################

################# do not edit below #######

# scriptname: server.pl

# this socket_serverscript server.pl should run to BSD, Illumos/Solaris, Linux, OSX and Windows
# for remote management from a Windows client with the cs web-gui

# start the ZFS backend server part:
# copy the folder ./[web-gui]/data/remote_cs_server to any location on a supported platform ex to /tank/data
# and start cs-backend server (as admin/root) via perl /path_to/remote_cs_server/start_server_as_admin.pl

# Add this server to a socketgroup in the frontend client Windows ZFS gui in menu Extensions > Socketgroup
# the auth key here in /socket/server.auth must be the same as on Windows in /path/_log/groups/hostname~ip.txt


#lib and path are set by startscript
use lib "/xampp/web-gui/data/cs_server/CGI";

# path to serverscript
$spath="/xampp/web-gui/data/cs_server";

#tmp folder cache files on remote server
$cpath="$spath/tmp";                         # cachepath

# localhost tmp => /xampp/web-gui/tmp  (localhost cache files)
if ($spath=~/\/xampp/i) {
   $cpath="/xampp/web-gui/tmp";              # cachepath
}

#remote.log for background tasks
$rl="$spath/tmp/remote.log";

#command.log
$cl="$spath/tmp/command.log";

#monitor.log
$ml="$spath/tmp/monitor.log";

if (!($max_connections > 0)) { $max_connections = 20; }      # defaults
if (!($timeout > 0))         { $timeout = 90; }

print "\ntimeout:$timeout\nmax_connections:$max_connections\nallowed_cmd:$allowed_cmd\nallowed_ip:$allowed_ip";

use IO::Socket::INET;                                        # socket communication
use IO::Select;

use HTTP::Tiny;                                              # send large data >32K via post
use URI::Escape;                                             # url escape ex " " -> %20

# other modules
#use File::Stat;                                              # file stat like age
use File::Find;                                              # find files or directories

use Fcntl qw(:flock);                                        # filelocking
use IPC::Run qw( run timeout );                              # system() alternative with stdout, stderr and timeout

use Digest::SHA::PurePerl qw(sha256_hex);                    # replaces md5m, $hash = sha256_hex($data); 96 char

use utf8;

# auto-flush on socket
$| = 1;

# Configurable parameters
#################################################

# set in config.pl
#my $allowed_cmd="";                                          # which console apps are allowed beside zfs and zpool
#my $allowed_ip="";                                           # which clients are allowed (commalist possible), ex 192.168.1.3,172.16.1.
                                                              # client ip must begin with these values or is denied
#my $max_connections="20";
#my $timeout="30"

my $port = 63000;
my $chunk_size = 16 * 1024;                                      # only when using chunks
                                                              # a socket transfer is limited to 32K to avoid buffer problems
# main #
############# some infos #########################
print "\n" x 15;                   # some newlines
print "cs_server $ver\n\n";


#################################
print "###########################\n";
print "initiate cs_server $ver\n";
#################################


# read ps and base infos #########################
##################################################

 &get_pw;         # print "pw=$data{'myauth'}\n"; # debug
 &my_os;          # detect os,on,hn,rh, path and print OS info

&socket_inet();   # start socket server
exit;

#####################################################################
sub socket_inet {      #ok, socket via Socket::INET (Socket::IP has problems)
################
   #
   print "use IO::Socket::INET\n";

   # intiate socket server
   my $server = IO::Socket::INET->new(
       LocalPort => $port,
       Proto     => 'tcp',
       Listen    => $max_connections,
       Blocking  => 0,  # Set to non-blocking mode
   ) or die "Cannot create socket: $!";

  my $select = IO::Select->new($server);

  while (1) {

       foreach my $client ($select->can_read($timeout)) {
           if ($client == $server) {
               my $new_client = $server->accept;
               $select->add($new_client);

               my $ip=$new_client->peerhost;

           } else {

               my $bytes_read = $client->sysread(my $message, $chunk_size);
               if ($bytes_read) {

                   my $ip=$client->peerhost;
                   if (&client_ip_allowed($ip) eq "0")  {

                          print __LINE__." client $ip not allowed\n";
                          $client->syswrite("your ip is not allowed\n");

                          $select->remove($client);
                          $client->close;

                   } else {

                      #  client is allowed
                      my @chunks;
                      my ($reply,$filename) = &process_command($message, $ip);

                      if ($filename=~/\w/) {
                        $t="upload: $filename";
                      } else {
                        if (length($reply) <= 250) {
                          $t=$reply;
                        } else {
                          $t=substr("$reply $filename",0,250)." ..";  $t=~s/^ +//;
                        }

                        if ($reply=~/:BoD:/) {
                          if ($reply=~/:EoD:/s) { $t.=" :EoD:"; } else { $t.="!! no EoD marker !!"; }
                        }
                      }

                      $rsize=length("$reply$filename");
                      print "\n".__LINE__." #####################\nsocketserver_in:\n########################\n\n$message\n\n############### reply $rsize Bytes #####\n$t\n###############\n";

                      if ($filename=~/\w/) {           # 192.._abc123.data or upload error:

                          #print __LINE__." uploaded data to file $filename\n";
                          $client->syswrite("$filename\n");    # return filename or upload error


                      } else {            # return data

                          my $l=length($reply) +5;          # + end of send :EoS:

                          if ($l < 16*1024) {
                              $t="$reply\:EoS:\n";
                              $client->syswrite($t);        # endofsend
                              print __LINE__." transferred $l Bytes via single syswrite\n\n";

                          } else {       # return chunks

                              $t=0;
                              while (length($reply) > 0) {
                                 my $chunk = substr($reply, 0, $chunk_size, '');  # Entferne den Chunk aus den Daten
                                 $client->syswrite("$chunk");  # Schreibe den Chunk auf den Socket
                                 $t=$t+length($chunk);
                              }
                              print __LINE__." transferred $t Bytes in $chunk_size Byte chunks\n\n";
                          }
                      }

                      $select->remove($client);
                      $client->close;
                  }

               } else {

                   print "\n#####################\n".__LINE__." Connection closed for " . $client->peerhost . "\n";
                   print "cs_server $ver\n\n";

                   $select->remove($client);
                   $client->close;
               }
           }
       }
   }

   # end of loop
   $server->close;

}  #/inet


######################
sub process_command {
######################
  # a request is either from 127.0.0.1 (Windows frontend to itself, unresticted) or

  # remote server call from frontend with ip
  # stat commands are processed directly. All other commands and results can be "direct" or "callback" type vih https (upsrv defined)
  # cmd for callback  rcmd::filename.rcmd (::options::)

  my ($message, $ip) = @_;
  my $q=$message;                                               # ip is calling host
  my ($r);

  #get options -> %data

  &parameter($message,$ip);                                     # -> %data, (::line::member::upsrv::w2::action::opt::filename::data::cache::time::cfn::auth::)
  $q=~s/ *\(::.*//s;                                            # remove cmd options

  # auth and allowed ip check, not localhost
  if ($ip ne "127.0.0.1")   {                                   # local check or always allowed?
    if ($data{'auth'} ne $data{'myauth'})    {                  # p(2)=auth
       if ($data{'myauth'}=~/missing/) { return("error: server.auth not found"); }
       print __LINE__." myauth: $data{'myauth'} ne auth\n $data{'auth'}\n";             # debug
       return('error: socket auth wrong');
    }

    # client_ip_allowed is already checked
  }
  ###
  $q=~s/\s+$//s;

  if ($q eq "") {                                              # error
       print __LINE__." <- [no cmd]\n";
       return ("error: [no cmd]");
  }
                                    # insert auth to curl cmd
  #print __LINE__." process command> $q\n------\n";              # see print ipc>

  # server runs in a cmd shell
  # commands on Windows are mostly using powershell
  # to execute powershell commands ex get-disk
  # you must execute: powershell -command get-disk
  # ex @q = qw(powershell -command Get-PhysicalDisk | Select DeviceID,UniqueId);

  # always use powershell, otherwise replace

  # short status commands always allowed after auth and ip check
  if (!($q=~/^hostname|^hn|^rh|^on|^hi|^os|^dn|^ip|^fn|^alive|^mem|^sc/)) {

    # too old, del
    $t=&my_age();      # age of activity.log max 7 days
    if ($t > 3600*24*7) {
      unlink ($cl);
      unlink ($rl);
    }

    $t=&get_mydate;
    &my_add($cl,"$t \t$q");                # add q to command.log
  }  else {
    print __LINE__." in> $q\n";
  }

  my $callback;
  if ($data{'enforce_upsrv'} ne "") { $callback=",callback"; }                                           # enforce callback

  # these short cmds always processed directly after auth verification
  if ($q eq "hostname") { print __LINE__."\n ok: <- $hn\n";         return ($hn);           }            # hostname
  if ($q eq "hn")       { print __LINE__."\n ok: <- $hn\n";         return ($hn);           }            # hostname
  if ($q eq "mem")      { print __LINE__."\n ok: <- $mem\n";        return ($mem);          }            # member ip
  if ($q eq "rh")       { print __LINE__."\n ok: <- $rh\n";         return ($rh);           }            # remote host
  if ($q eq "hi")       { print __LINE__."\n ok: <- $on,$rh,$os\n"; return ("$fn\;$dn\;$rh\;$os\;cs $ver$callback"); }  # remote host info family +  remote host + os
  if ($q eq "os")       { print __LINE__."\n ok: <- $os\n";         return ($os);           }            # os
  if ($q eq "sc")       { print __LINE__."\n ok: <- $sc\n";         return ($sc);           }            # samba config
  if ($q eq "on")       { print __LINE__."\n ok: <- $on,$rh,$os\n"; return ("$fn\;$dn\;$rh\;$os\;cs $ver$callback"); }   # on same as hi, opt callback
  if ($q eq "dn")       { print __LINE__."\n ok: <- $dn\n";         return ($dn);           }            # distri name
  if ($q eq "fn")       { print __LINE__."\n ok: <- $fn\n";         return ($fn);           }            # family name
  if ($q eq "alive")    { print __LINE__."\n ok: <- yes,alive\n";   return ("yes,alive");   }            # alive?
  if ($q eq "ip")       { print __LINE__."\n ok: <- $ip\n";         return ($ip);           }            # ip of client


  #################  end of direct actions beside localhost, others may need callback ##


  if ($ip ne "127.0.0.1") {             # can be a direct call or download callback request
  #########################

      # if $up eq "", no callback $cmd  or up is enforced locally or can be requested from frontend  -> callback (download cmd)
      # result is always uploaded

      # download cmd?
      if (($data{'upsrv'} ne "") || ($q=~/^rcmd::/))  {           # callback rcmd::filename (::option::)
                                                       # cmd for callback  rcmd::upsrv::cfn::memberip::auth:::opt(nocachefrontendtime)::
          $q=&my_download("$data{'upsrv'}","$q","$data{'upauth'}",$data{'ip'});          # rcmd::filename
          if ($q=~/error:/s) { return ($q,""); }       # reply,filename
      }

      &parameter($message,$ip);                                     # reload with up cmd -> %data
      $q=~s/ *\(::.*//s;                                            # remove cmd options

      $t=substr($q,0,20); $t=~s/\n/<n>/gs;
      print __LINE__." remote server> $t ..\n";

      # now we have a valid command either direct or via callback

  }

  ###########################################################
  ## this is localhost execution or after .rcmd download ##
  ############################################################

  my $ps="1";                        # use powershell
  my $grep="";                       # simualate a basic grep filter with grep a or grep a,b (a and b)

  if ($q=~/\| *grep/) {              # ex tasklist | grep perl   or grep perl,nc   or grep 1234,1235
       $grep="$q";
       $grep=~s/.*grep +//;          # get filter
       $q=~s/ *\| *grep.*//;         # remove grep
       print __LINE__." grep filter: $grep\n";
  }

  # check auth
  if ($data{'myauth'} ne $data{'auth'})    {
       print __LINE__." pw: $data{'myauth'} ne auth $data{'auth'}\n";       # debug
       if ($data{'myauth'}=~/missing/) { return("error: server.auth not found"); }
       return('error: auth error');
  }
  ################# process cmd ####################

  # background task ex zfs send snap | nc ip port &
  if ($q =~/\&$/) {

     $r=&background($q);
     return ($r);                     # diff pid

  } else {

     if ($on=~/mswin/i)  {              # windows, not darWIN

       if ($q=~/^type /)  { $q=~s/^type /get-content /;    }    # replace with powershell command
       if ($q=~/^ps axw/) { $q=~s/^ps axw */tasklist/; $ps=0; } # replace with cmd command
       if ($q=~/^curl/)   { $ps=0; }                             # curl in powershell ?

       @q=();

       if  ($ps) {                                                  # use powershell
           #
           if ($q=~/^curl/) {
                  $t=$q;
                  $t=~s/^curl *//;
                  # curl --connect -- connect.pl
                  # Start-Process -FilePath 'curl' -ArgumentList 'curl --connect-timeout 3 -X POST -k -F "upfile=123.rcmd" -F "auth=345" -F "data=YOUR_DATA_PARAMETER" https://localhost/cgi-bin/cs/cs_connect.pl' -NoNewWindow
                  $q[0]="powershell";
                  $q[1]="Start-Process";
                  $q[2]="$t";              # curl parameter
                  $q[3]="-NoNewWindow";

           } else {
                #
                $q=~s/^powershell +|^ps +//;                         # no double powershell  (from client)
                $q=~s/^ *\-command +//;                              # no double -command
                $q[0]="powershell";                                  # use array for irc::run
                $q[1]="-command";

                $q[2]="$q";                                          # remove opt
            }

          } else {

             @q=();
             @q=split(/ /,$q);                                    # use cmd
          }

     } else {  #/not win

       @q=();                                                    # crash in omnios with replacements on modify call
       @q=split(/ /,$q);
     }

  }  # /background task                                          # powershell command
  #

  if ($q eq "stat")     { print __LINE__."\n ok: <- todo\n";      return ('todo');      }            # basic stat
  my $out1="";

  if ($q eq "rl")       {
      print __LINE__."\n ok: rl <- remote.log\n";
      $out1=&my_read("$rl");
      if (!($out1 =~/\w/s)) {
        $out1='empty';
      }
  }

  if ($q eq "cl")       {
      print __LINE__."\n ok: cl <- command.log\n";
      $out1=&my_read("$cl");
      if (!($out1 =~/\w/s)) {
        $out1='empty';
      }
  }                # command log
  if ($q eq "rld")      { print __LINE__."\n ok: rld <- remote.log del\n";  $out1=&my_read("$rl");  unlink($rl);  }   # remote log del
  if ($q eq "cld")      { print __LINE__."\n ok: cld <- command.log del\n"; $out1=&my_read("$cl");  unlink($cl);  }   # command log del
  if ($q eq "ml")       { print __LINE__."\n ok: ml <- monitor.log\n";      $out1=&my_read("$ml");   }                # monitor log
  if ($q eq "mld")      { print __LINE__."\n ok: mld <- monitor.log del\n"; $out1=&my_read("$ml");  unlink($ml);  }   # monitor log del


  # unlock, #append or write
  if ($q=~/^rcmd::writefile/)    { $t=&write_file($q);   print "write:$t\n";  return($t); }           # writefile  ex /etc/samba/smb.conf
  if ($q=~/^rcmd::appendfile/)   { $t=&append_file($q);  print "append:$t\n"; return($t); }           # appendfile ex /etc/pam.conf
  if ($q=~/^rcmd::unlock/)       { $t=&unlock($q);       print "unlock:$t\n"; return($t); }           # unlock fs
  if ($q=~/^rcmd::del_tmp_keys/) { $t=&del_tmp_keys($q); print "del tmp_keys:$t\n"; return('deleted ./tmp/*.completekey'); }           # unlock fs

  if (($q=~/zfs |zpool /) && ($haszfs eq "no"))  {
         print __LINE__."\n ok: <- zfs not installed\n";
         return ("zfs not installed");            # zfs not installed
  }

  #
  # filename for upload data

  my $filename="$data{'cfn'}";          # 127.0.0.1_zpool_list..
  my $lfilename=$filename;

  $lfilename.=".last";
  $filename.=".data";                #  local cachefile

  # del cache
  if ($q eq "dc")       {
     my @f = glob("$cpath/*.*");     # cachepath
     foreach my $f (@f) {
        if (!($f=~/\.data$|\.last$/)) {  next; }
        unlink($f);
     }
     print __LINE__."\n ok: <- dc\n";        return ('ok, cache files deleted');
  }


  if ((-f "$cpath/$filename") && ($nocache ne "1")) {           # use cache file

       #check last (time()\fullcmd)

       my @last=&my_read($lfilename,'a');      # time\ncmd
       $tm=time();

       if ($last[0]-$tm < 300) {               # cache 5min
             print __LINE__." could use cachefile, todo: $filename\n";
             #$out1=&my_read($filename);
       } else {
             unlink($filename);
             unlink($lfilename);
       }
  }

  if ($nocache) { print __LINE__." cache disabled\n"; }         # this is a command option: cmd opt:=nochache auth:=123

  # keylist
  ###########
  if ($q eq "keylist") {
     $t=&my_keylist($ip);
     return($t,"");
  }

  # run cmd -> file as size of $out is limited, return only error on small io directly
    my $err="";
    my $out="";
    my $in="";

  $t=join(" ",@q);
  print __LINE__." ipc run> $t $in\n-------------\n";             # opt, socket in is usually enough

  if ($out1 eq "") {           # no log or cache
    ###############################################
    eval {
      if ($q[0] eq "curl") {                 # problems with ipc<..run ?
         $out=`$t`
      } else {
        run \@q, \$in, \$out, \$err, timeout( $timeout );          # execute via ipc::run and timeout (in s)
      }
    };
    ################################################
  }   else {
     $out=$out1;  # log or cache files
  }
  #;

  if ($q=~/^wmic /) {                # win replacement of ps axw, gives special char
      $out=~s/[^a-zA-Z0-9\n\t\r \.,;\-\_\#]//gs;
  }

  if ($@) {
      # eval error
      $err=$@;
      print __LINE__." eval error: $@\n";
  }


  $out=~s/\n+$//s;             # no trailing nl
  $err=~s/\n+$//s;             # no trailing nl

  if ($err =~/error:/) {

           print __LINE__."\n <- error: \n$err\n";
           return ("error: $err");  # stderr

  } else {

           # simulate basic grep
           if ($grep ne "") {
                my $ok;
                my @greps=split(/,/,$grep);                           # allow commalist based on and
                print __LINE__." grep $grep\n";

                my @out=split(/\n/,$out);
                my @filtered=();
                foreach my $o  (@out) {

                  $ok="";
                  foreach my $g (@greps) {                            # grep as commalist a,b -> a and b
                    #print "grep $o -> $g\n";
                    if (!($o=~/$g/)) { $ok.="0";  }                    # not found
                  }

                  if (!($ok=~/0/))    {
                     push(@filtered,$o);                               #every grep in outpur line
                  }
                }
                $out=join("\n",@filtered);
           }                                                           # include err
           if ($err ne "") { $out.="\n$err"; }

           #utf-8 ?
           #utf8::encode($out);              # umlaut wrong at console

           #if ($t=~/^(?:[\x00-\x7F]|\xC0[\x80-\xBF]|\xE0[\x80-\xBF][\x80-\xBF]|\xF0[\x80-\xBF][\x80-\xBF][\x80-\xBF])*$/) {
           #  print __LINE__." likely utf-8\n";
           #} else {
           #  print __LINE__." propably not utf-8\n";
           #}

           # save result as file (caching and file upload
           $rsize=int(length($out)/1024);                        # result size
           if ($rsize < 1) { $rsize=1; }
           $rsize.="K";

           # write filename.data for caching
           &my_write("$cpath/$filename",$out);                    # cfn.data

           # write lfilename.last for caching
           $tm=time();
           &my_write("$cpath/$lfilename","$tm\n$q\n");            # cfn.last

           # upload, never on localhost (web-gui can read tmp folder)
           # for a remote server upload when $upsrv is defined or whwn size > 8K

           if ($ip ne "127.0.0.1")  {
              print __LINE__." up=$data{'upsrv'}\n";

              if (($data{'upsrv'} ne "") || (length($out) > 1024*60)) {                  # 4k, chunked transfer, upload if larger 4MB or forced by upsrv
                   #                                                                     # socket with problems on larger transfers
                   $r="";
                   if ($data{'upsrv'} eq "") { $data{'upsrv'}=$ip; } # fallback option
                   eval {
                      $r=&my_upload($data{'upsrv'},$filename,$data{'upauth'},$ip);       # upload file via  curl: up, filename, auth, ip
                   };

                   if ($@) {
                           #   eval error
                           $err=$@;
                           print __LINE__." eval error: $@\n";
                   }

                   ## return filename  or error from my_upload:
                   if ($r=~/error:/s)  {
                          return($r,"");
                   } else {
                          return("",$filename);
                   }

              } else {      # return data directly with EoD marker to detect partly transmissions

                    $t=substr($out,0,700);  $t=~s/\n| +/ /gs;
                    print __LINE__." return $rsize data:\n$t ...\n";
                    return(":BoD:$out\:EoD:","");                # with BoD/EoD Beginofdata, endofdata marker to detect incomplete transfers
              }

           } else {

              #localhost, return data or filename

              if (length($out) > 1024*60) {             # 4 MB
                 return("",$filename);
              } else {
                 return(":BoD:$out\:EoD:",'');
              }
           }
  }
}

#############
sub parameter {  # command with parameter , cmd can have :: in rcmd::writefile
#############     cmd (:: -> line::member::upsrv::w2::action::opt::filename::data::cache::time::cfn::auth::upauth)
               #  rcmd::unlock (::..

   my $cmd=$_[0];                          # can have :: in cmd
   my $opt=$cmd; $opt=~s/.*\(:://;         # unlink::fs:: (::line::..


   my $ip=$_[1];      # calling ip

   $data{'ip'}    =$ip;

   $cmd=~s/\n|\r//gs;
   $cmd=~s/^ +| +$//g;

   my ($t,$p);

   #                                          0     1       2      3     4     5
   if ($cmd=~/\(::/) {     # rcmd::aa::bb  (::line  member  upsrv  w1..

       $opt="cmd::$opt";                       # data up from pos1
       my @par=split(/::/,$opt);

       # cmd=par[0]
       $data{'line'}    =$par[1];
       $data{'member'}  =$par[2];
       $data{'upsrv'}   =$par[3];
       $data{'w1'}      =$par[3];
       if ($data{'w1'} eq "") { $data{'w1'}=$ip; }
       $data{'w2'}      =$par[4];
       $data{'action'}  =$par[5];
       $data{'opt'}     =$par[6];
       $data{'filename'}=$par[7];
       $data{'data'}    =$par[8];
       $data{'cache'}   =$par[9];            # "nocache"
       $data{'time'}    =$par[10];           # frontend time 12345
       $data{'cfn'}     =$par[11];
       $data{'auth'}    =$par[12];           # auth web-gui -> this server
       $data{'upauth'}  =$par[13];           # auth this server -> up

       #enforce local settings
       if ($data{'_upsrv'} ne "")      { $data{'upsrv'}=$data{'_upsrv'}; }
       if ($data{'_upsrv_auth'} ne "") { $data{'upsrv_auth'}=$data{'_upsrv_auth'}; }
       if ($data{'_w1'} ne "")         { $data{'w1'}=$data{'_w1'}; }
       if ($data{'_w1_auth'} ne "")    { $data{'w1_auth'}=$data{'_w1_auth'}; }
       if ($data{'_w2'} ne "")         { $data{'w1'}=$data{'_w1'}; }
       if ($data{'_w2_auth'} ne "")    { $data{'w2_auth'}=$data{'_w2_auth'}; }
   }

   if ($data{'cache'} =~/delcache/)  {

           #del cache
           my @files = glob("$cpath/*.*");
           foreach my $f (@files) {
              if ($f=~/\.data/) {               # only data files, all on start
                 unlink ($f);
              }
           }

   }

   # callback
   if ($data{'upsrv'} eq "ip") {
       $data{'upsrv'}=$ip;                    # remote cmd can be "", 'ip' or 'hostname.dns'
   }

   #$data{'auth'}=~s/[^a-f0-9]//g;            # hex only?
   print "\n".__LINE__." sub parameter:\nline=$data{'line'}, member=$data{'member'}, upsrv=$data{'upsrv'}, w2=$data{'w2'}, action= $data{'action'}, opt=$data{'opt'}, filename=$data{'filename'}, data=$data{'data'}, cache=$data{'cache'}, time=$data{'time'}, cfn=$data{'cfn'}, auth=$data{'auth'}, upauth=$data{'upauth'}\n\n";

   return();   # -> %data
}

###
sub my_keylist {                   # return list of all available keys
###

my $ip=$_[1];

# get a list of all key(parts) accessable by this server
my $member=$data{'member'};              # member name~ip

# local keys
#############
my $kpath="$spath/keys";                 # local keypath
if ($kpath=~/xampp/) {                   # win
    $kpath="/xampp/web-gui/_log/keys";
}

my @l_keys = glob("$kpath/*.*");         # all local keys
foreach my $k (@l_keys) {
   $k=~s/.*\///;                         # without path
}
my $l_keys="local ($member) [$kpath]\n".join("\n",@l_keys);   # commalist

# webserver1
##############                           # win apache or remote https w1 (settings top of script)
my $w1=$data{'w1'};
if ($w1 eq "")               { $w1=$ip; }                          # web-gui
if ($data{'_w1'} ne "")      { $w1=$data{'_w1'}; }                 # enforce local

my $w1_auth=$data{'upauth'};
if ($data{'_w1_auth'} ne "") { $w1_auth=$data{'_w1_auth'}; }       # enforce local

#get keys
my $k=" -k";                      # allow private cert
if ($w1=~/[a-z]/) { $k=""; }   # no self signed certs

my $w1_keys="w1 ($w1)".`curl --connect-timeout 4$k -X POST -F "getfile=$member.keys" -F "auth=$w1_auth" https://$w1/cgi-bin/cs/cs_connect.pl`;
chomp($w1_keys);
print __LINE__."w1 ($w1) curl --connect-timeout 4$k -X POST -F \"getfile=$member.keys\" -F \"auth=$w1_auth\" https://$w1/cgi-bin/cs/cs_connect.pl\n$w1_keys\n";


# webserver2                             # remote https w2 (settings top of script)
#############
my $w2=$data{'w2'};
if ($data{'_w2'} ne "")      { $w2=$data{'_w2'}; }                 # enforce local
my $w2_auth="";
my $w2_keys="";

if ($w2 ne "") {

     $w2_auth=$data{'upauth'};
     if ($data{'_w2_auth'} ne "") { $w2_auth=$data{'_w2_auth'}; }  # enforce local

     #get keys
     my $k=" -k";                      # allow private cert
     if ($w2=~/[a-z]/) { $k=""; }      # no self signed certs

     $w2_keys="w2 ($w2)".`curl --connect-timeout 4$k -X POST -F "getfile=$member.keys" -F "auth=$w2_auth" https://$w1/cgi-bin/cs/cs_connect.pl`;
     chomp($w2_keys);
     print __LINE__."w2 ($w2) curl --connect-timeout 4$k -X POST -F \"getfile=$member.keys\" -F \"auth=$w2_auth\" https://$w2/cgi-bin/cs/cs_connect.pl\n$w2_keys\n";

} else {
     $w2_keys="w2 (not set) [-]";                      #(member) [path]
}

return ("$l_keys;\n$w1_keys;\n$w2_keys");             # all known keys


}  #/sub

############ rcmd::unlock::fs(pool/data)::hash::
sub unlock { #            0        1        2       3       4              5
############
   my ($r,$k,@t);
   my $q=$_[0];
   $q=~s/^rcmd:+//;

   print "\n".__LINE__." %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\nrcmd:unlock> $q\n%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n\n";
   my $ip=$_[1];          # default source of missing keyparts
                          # per default 1/3 of key is local
                          # 2/3 and 3/3 are on web-gui (or w1,w2)
                          # you either can use 3 sources (l:w1:w2) or 2 sourcs in ha (l:w*)
                          # upserver and upauth can be enforced in cs_server settings

   my @p=split(/::/,$q);
   my @key;                  # 0=fs, 1-3=keyparts, 4=compl
   my $kpath="$spath/keys";  # local keypath

   # my @keys = glob("$kpath/*.*");      # all local keys ?

   # needed
   my $member=$data{'member'};
   my $filesystem=$p[1];     # ex pool/data

   my $tfilesystem=$filesystem;
   $tfilesystem=~s/\//~/g;   # filename with ~ instead /

   $key[0]=$filesystem;
   my $hash=$p[2];           # ''(file), pw, hash

   if (($member eq "") || ($filesystem eq "")) { return("error: need member and filesystem"); }

   # no hash: search parts
   if (($hash eq "") || ($hash=~/~/))    {  # ./tmp/omnio46~192.168.2.44~tank~enc.completekey

       # local keyparts?
       if (-f "$kpath/$member\~$tfilesystem.completekey")              { $key[4]=&my_read("$kpath/$member\~$tfilesystem.completekey"); $key[4]=~s/[^a-z0-9]//gs; }  # unusual
       if ($key[4] eq "") {
           if (-f "$kpath/$member\~$tfilesystem.keypart1")             { $key[1]=&my_read("$kpath/$member\~$tfilesystem.keypart1"); $key[1]=~s/[^a-z0-9]//gs;   }
           if (-f "$kpath/$member\~$tfilesystem.keypart2")             { $key[2]=&my_read("$kpath/$member\~$tfilesystem.keypart2"); $key[2]=~s/[^a-z0-9]//gs;   }
           if (-f "$kpath/$member\~$tfilesystem.keypart3")             { $key[3]=&my_read("$kpath/$member\~$tfilesystem.keypart3"); $key[3]=~s/[^a-z0-9]//gs;   }
           if (($key[1] ne "") && ($key[2] ne "") && ($key[3] ne ""))  { $key[4]="$key[1]$key[2]$key[3]"; }
       }

       #debug
       #print __LINE__." local keys: $key[0] ::$key[1] ::$key[2] ::$key[3] ::$key[4]\n";

       ########
       # try w1
       ########

       my $w1=$data{'w1'};
       if ($w1 eq "")               { $w1=$ip; }                          # web-gui
       if ($data{'_w1'} ne "")      { $w1=$data{'_w1'}; }                 # enforce local

       my $w1_auth=$data{'upauth'};
       if ($data{'_w1_auth'} ne "") { $w1_auth=$data{'_w1_auth'}; }       # enforce local

       #get keys
       $k=" -k";                      # allow private cert
       if ($w1=~/[a-z]/) { $k=""; }   # no self signed certs

       $r=`curl --connect-timeout 4$k -X POST -F "getfile=$member\~$tfilesystem.keys" -F "auth=$w1_auth" https://$w1/cgi-bin/cs/cs_connect.pl`;
       chomp($r);
       print __LINE__."w1: curl --connect-timeout 4$k -X POST -F \"getfile=$member\~$tfilesystem.keys\" -F \"auth=$w1_auth\" https://$w1/cgi-bin/cs/cs_connect.pl\n$r\n";

       if ($r=~/::/)  {          # valid data: fs::kp1::kp2::kp3::compl
           @t=split(/::/,$r);
           if (($key[1] eq "") && ($t[1] ne "")) { $key[1]=$t[1]; }
           if (($key[2] eq "") && ($t[2] ne "")) { $key[2]=$t[2]; }
           if (($key[3] eq "") && ($t[3] ne "")) { $key[3]=$t[3]; }
       }
       print __LINE__." w1 keys($w1): $t[0] ::$t[1] ::$t[2] ::$t[3] ::$t[4]\n";
       if (($key[1] ne "") && ($key[2] ne "") && ($key[3] ne "")) { $key[4]="$key[1]$key[2]$key[3]"; }

       ########
       # try w2
       ########

       if ($key[4] eq "")   {
          my $w2=$data{'w2'};
          if ($data{'_w2'} ne "") { $w2=$data{'_w2'}; }                 # enforce local

          if ($w2 ne "") {

                     my $w2_auth=$data{'upauth'};
                     if ($data{'_w2_auth'} ne "") { $w2_auth=$data{'_w2_auth'}; }  # enforce local

                     #get keys
                     $k=" -k";                     # allow private cert
                     if ($w1=~/[a-z]/) { $k=""; }   # hostname, no self signed certs

                     $r=`curl --connect-timeout 4$k -X POST -F "getfile=$member\~$tfilesystem.keys" -F "auth=$w2_auth" https://$w2/cgi-bin/cs/cs_connect.pl`;                     chomp($r);
                     print __LINE__."w2: curl --connect-timeout 4$k -X POST -F \"getfile=$member\~$tfilesystem.keys\" -F \"auth=$w2_auth\" https://$w2/cgi-bin/cs/cs_connect.pl\n$r\n";

                     if ($r=~/::/)  {          # valid data: fs::kp1::kp2::kp3::compl
                         @t=split(/::/,$r);
                         if (($key[1] eq "") && ($t[1] ne "")) { $key[1]=$t[1]; }
                         if (($key[2] eq "") && ($t[2] ne "")) { $key[2]=$t[2]; }
                         if (($key[3] eq "") && ($t[3] ne "")) { $key[3]=$t[3]; }
                     }
                     print __LINE__." w2 keys($w2): $t[0] ::$t[1] ::$t[2] ::$t[3] ::$t[4]\n";

                     if (($key[1] ne "") && ($key[2] ne "") && ($key[3] ne "")) { $key[4]="$key[1]$key[2]$key[3]"; }

                     if ($key[4] eq "") { return ("error: could not find key or all keyparts."); }
           }
       }

       $hash=$key[4];

   } else {          # hash (can be sha256hex pw or real key)

          $t=`zfs get -H -o value keylocation $filesystem`;
          print __LINE__." keylocation $filesystem: $t\n";
          if ((length($hash) <= 32) && (!($t=~/prompt/)) ) {                 # pw hash
              print __LINE__." create hash from fs+pw\n";

              $t="$filesystem$hash";
              $t=~s/~|\///g;                       # hash from filesyszem + pw ex tankdatageheim
              $hash= sha256_hex($t);               # create key as hash from pw
          }
   }

   #try load key $hash
   if ($hash eq "") { return("error: sub unlock, no key for $member\~$filesystem"); }

   if ($on =~/Solaris 11|MSwin/i) {       # unlock via tmp file

      $t="$cpath/$member~$tfilesystem.completekey";                     # cachepath
      &my_write("$cpath/$member~$tfilesystem.completekey",$hash);

      if ($on =~/Solaris 11/i) {

         print __LINE__." zfs key -l $filesystem\n";
         print `zfs key -l $filesystem`;

      } else {

          print __LINE__." zfs load-key -r $filesystem\n";
          `zfs load-key -r $filesystem`;
      }
      unlink("$cpath/$member~$tfilesystem.completekey");         # only temp file to unlock

   }  else {        # unlock Open-zfs (Windows has problems) via echo, no file needed

      $hash=~s/\s//gs;

      $t=substr($hash,0,7)."..";
      print "--unlock--\n";
      print __LINE__." echo $t | zfs load-key -L prompt $filesystem\n";
      #print __LINE__." echo $hash | zfs load-key -L prompt $filesystem\n\n";    # debug

      `echo $hash | zfs load-key -L prompt $filesystem`;
   }

   $r=`zfs get -H keystatus $filesystem`."\n";
   print __LINE__." keystatus $filesystem: $r\n";

   if ($r=~/\savailable/)   {
        return('available');
   } else {
        return('unavailable');
   }

}  #/sub

##############
sub append_file {
##############
# remote request to append data to file ex /etc/pamconf
# command: rcmd:appendfile::file::data::EndOfFile::

my $data=$_[0];
my @parts=split(/::/,$data);
if ($parts[0] ne "rcmd:appendfile")       {  return ("error: no appendfile command"); }
if (!($parts[1]=~/\/\w|\\\w/))            {  return ("error: no path+filename"); }
#if (!($parts[1]=~/.conf/)) {  return ("error: file not allowed"); }   # smb.conf or smb4.conf
if ($parts[2] eq "")                      {  return ("error: nodata"); }
if ($parts[3] ne "EndOfFile")             {  return ("error: EndOfFile not found"); }

#keep default
if (!(-f "$parts[1].default")) {
  `cp $parts[1] $parts[1].default`;           # ex pam.conf.default
}

# single line transfer
$parts[2]=~s/<n>/\n/gs;                       # single line transfer

# edit in /tmp
print __LINE__." cp $parts[1] /tmp/append.tmp\n";
`cp $parts[1] /tmp/append.tmp`;

print __LINE__." appendfile $parts[2]\n";

&my_append("/tmp/append.tmp","$parts[2]");    # add tmp file
`cp /tmp/append.tmp $parts[1]`;               # copy back
`rm /tmp/append.tmp`;                         # rm tmp
return ("ok");
}

##############                                                                                              lob
sub del_tmp_keys {
##############
    # remote request to del all complete keys in tmp folder
    # command: rcmd::del_tmp_keys::

    my $data=$_[0];
    my @parts=split(/::/,$data);

    if ($parts[1] ne "del_tmp_keys")          {  return ("error: no del_tmp_keys command"); }


     my @f = glob("$spath/tmp/*.*");     # tpath
     foreach my $f (@f) {
        if (!($f=~/\.completekey$/)) {  next; }
        unlink($f);
     }

}

##############
sub write_file {
##############
# remote request to write file ex /etc/samba/smb.conf
# command: rcmd::writefile::file::data::EndOfFile::

# or keyfile  member~pool~fs.completekey

my $data=$_[0];
my @parts=split(/::/,$data);

if ($parts[1] ne "writefile")             {  return ("error: no writefile command"); }
if ($parts[2] eq "")                      {  return ("error: no file"); }
if ($parts[3] eq "")                      {  return ("error: nodata"); }
if ($parts[4] ne "EndOfFile")             {  return ("error: EndOfFile not found"); }

# allowed file extensions
my $ext=$parts[2];
if (!($ext=~/\.\w+/))  { return ('writefile: files must have an extension like .conf'); }
$ext=~s/.*\.//;

if ($ext eq "conf") {
   if (!($parts[1]=~/\/\w|\\\w/))             {  return ("error: no path+filename"); }
   if (!($parts[1]=~/smb\.conf|smb4\.conf|/)) {  return ("error: file not allowed"); }   # smb.conf or smb4.conf
}

if ($ext eq "completekey") {               # keyname, add path
   $parts[2]="$cpath/$parts[2]";           # key in tmp folder
}

#keep default conf
if ( (!(-f "$parts[2].default")) && ($ext=~/conf/)) {
  `cp $parts[2] $parts[2].default`;        # smb.conf.default
}

# single line transfer
$parts[2]=~s/<n>/\n/gs;                    # single line transfer

print __LINE__." writefile $parts[2]\n$parts[3]\n$parts[4]\n";
&my_write("$parts[2]","$parts[3]");    # write file
return ("$parts[2]");                  # return filename

}

##################
sub get_smb_conf {
##################
   my ($r,$t);

   # search smb.conf or smb4.conf
   $r=`find /etc/samba -name smb.conf`;                            # default Linux
   chomp($r);
   if ($r=~/smb.conf/s) {
       $t=&my_read($r);
       if ($t=~/workgroup/s)  {   # valid config
            return($r);
       }
   }

   $r=`find /usr -name smb.conf`;                                  # TrueNas Linux, /usr/share/samba/smb.conf
   chomp($r);
   if ($r=~/smb.conf/s) {
       $t=&my_read($r);
       if ($t=~/workgroup/s)  {   # valid config
            return($r);
       }
   }

   $r=`find /usr/local -name smb4.conf`;                            # Free-BSD 14 /usr/local/etc/
   chomp($r);
   if ($r=~/smb4.conf/s) {
       $t=&my_read($r);
       if ($t=~/workgroup/s)  {   # valid config
            return($r);
       }
   }

   $r=`find /opt -name smb.conf`;                                  # omnios
   chomp($r);
   if ($r=~/smb.conf/s) {
       $t=&my_read($r);
       if ($t=~/workgroup/s)  {   # valid config
            return($r);
       }
   }
}

############
sub get_pw {
########### get pw hash

  my @data = ();
  $wpath=$spath;
  $wpath=~s/cs_server/cfg/;
  my $file="$wpath/server.auth";

  if ($^O=~/Win32/i) {
     $t=$spath; $t=~s/data.*/cfg/;     #/xampp/web-gui/cfg
     $file="$t/server.auth";
  }
  my (@p,$data);

  if (!(-f "$file")) { #same dir as server.pl
     print __LINE__." required $file not found\n";
     $data="cs_server_auth_missing";
     #exit;
  }

  print __LINE__." read auth: $file\n";
  $data=&my_read($file);

  # any string
  # $data="missing";    # debug

  $data=~s/.*value_adminpw/value_adminpw/s;      # old, now use simple string in server.auth
  $data=~s/\n.*//s; $data=~s/\s//gs;             # first line, no spaces

 # print "$data";
  $data{'myauth'}=$data;

  #del cache on start
  my @files = glob("$cpath/*.*");
  foreach my $f (@files) {
     #if ($f=~/\.data/) {               # all on start
        unlink ($f);
     #}
  }

  return(1);

}

################    command1 & or command1 | command2 &
sub background {
################

   my $command=$_[0];

   if (!($command =~/\&$/)) { print __LINE__." sub background: $command not a background job\n"; return; }

   $command=~s/ +$//;
   $command=~ s/\P{ASCII}//g;                    # remove non accii 127  char
   $command=~s/^l\_\d\d\d+ //;                   # line reference

   # ex for zfs send snap | nc ip 7000
   # Win: such a pipe does not work in powershell for unknown reason: start via cmd and start /B command

   ### file errorlog 2>&1 ?
   if ($on=~/mswin/i)  {                    # windows, not darWIN   / -> \
      $rl=~s/\//\\/g;
   }
   $command=~s/\&/ >> $rl 2>\&1 \&/;         # & -> >> remote.log 2>&1  &
   $command=~s/\bnc /$nc /;                 # must use nc from tools

   $t=&get_mydate;
   &my_add("$rl","$t\n");                   # date and time
   &my_add("$rl","\t$command\n");          #  command \n
   &my_add("$rl","re: \t$line \t");         # re 123  (return value of replications)

   my $cmd1=$command;
   my $cmd2="";
   my ($r,@p,$pids1,$pids2,@v,$id);

   if ($cmd1=~/\|/) {
     $cmd1=~s/ *\|.*//;
     $cmd1=~s/ .*//; # ex zfs send -> zfs

     $cmd2=$command;                     # after pipe
     $cmd2=~s/.*\| *//;
   }

   print __LINE__." start background task $command with cmds \n$cmd1 $cmd2\n";

   #get prozesses for cmd1 and cmd2 prior start
   if ($on=~/mswin/i)  {              # windows, not darWIN


      # Abbildname                     PID Sitzungsname       Sitz.-Nr. Speichernutzung
      # ========================= ======== ================ =========== ===============
      # System Idle Process              0 Services                   0             8 K
      # ..
      # services.exe                  1352 Services                   0        10.984 K
      eval {
        $r=`tasklist`;
      };
      @p=split(/\n/,$r);
      foreach my $p (@p) {
         @v=split(/ +/,$p);
         if ($v[0]=~/zfs|nc/) { $pids1.="$v[1],"; }            # check only nc and zfs
      }

   } else {

      #
      # ps axw
      # PID TT  STAT       TIME COMMAND
      # 557  -  Ss      0:00.95 /sbin/devd
      # 758  -  Ss      0:02.35 /usr/sbin/syslogd -s
       eval {
         $r=`ps axw | grep -v ps`;
       };
       @p=split(/\n/,$r);
       foreach my $p (@p) {
         $p=~s/^ +//;
         @v=split(/\s+/,$p);
         if ($v[4]=~/zfs|nc/) {         # check only zfs and nc
           $pids1.="$v[0],";
         }
       }
   }
   $pids1=~s/,+$//;                  # processes prior start of background job
   #print __LINE__." processes prior start: $pids1\n";

   eval {

      if ($on=~/mswin/i)  {            # windows, not darWIN
          $command=~s/\&$//;            # remove &
          print "\n".__LINE__." start background job: \n########################################################\n\n";
          #system("start /B $command");  # start in background via cmd -> always in background in new console

          #powershell -command "start-process C:\xampp\perl\bin\perl.exe cs_server\server.pl -WindowStyle minimized"

          print "powershell -command \"start-process \"cmd /c start $command\"\" -WindowStyle minimized\"\n";
          system("powershell -command \"start-process \"cmd /c start $command\"\" -WindowStyle minimized\"");

      } else {

        print "\n".__LINE__." start background job: \n############################\n\n$command\n\n############################\n\n";
        system("$command");          # has & at the end
      }
   };

   sleep(1); # wait for start
   print "ok, go on\n";

   #get prozesses for cmd1 and cmd2 after start
   if ($on=~/mswin/i)  {              # windows, not darWIN

      # Abbildname                     PID Sitzungsname       Sitz.-Nr. Speichernutzung
      # ========================= ======== ================ =========== ===============
      # System Idle Process              0 Services                   0             8 K
      # ..
      # services.exe                  1352 Services                   0        10.984 K
      eval {
        $r=`tasklist`;
      };
      @p=split(/\n/,$r);
      foreach my $p (@p) {
         @v=split(/ +/,$p);
         if ($v[0]=~/zfs|\bnc/) { $pids2.="$p,"; }
      }

   } else {
      #
      # ps axw
      # PID TT  STAT       TIME   COMMAND
      # 557  -  Ss      0:00.95   /sbin/devd
      # 758  -  Ss      0:02.35   /usr/sbin/syslogd -s

       eval {
         $r=`ps axw | grep -v ps`;
       };
       @p=split(/\n/,$r);
       foreach my $p (@p) {
         $p=~s/^ +//;
         @v=split(/ +/,$p);
         if ($v[4]=~/zfs|\bnc/) {         # check only zfs and nc
           $pids2.="$p,";
         }
       }
   }

   $pids2=~s/,+$//;                  # processes prior start of background job
  # print __LINE__."\nprocesses after start: $pids2\n";

   # get difference
   my $pid_new="";
   my @p2=split(/,/,$pids2);
   foreach my $v (@p2) {
       if ($pids1=~/\b$v\b/) {  next; } # pid was there prior start, we need only new ones
       $pid_new.="$v,";
   }

   $pid_new=~s/,+$//;
   print __LINE__." new processes after start: \n$pid_new\n\n";
   &my_add($rl,"new processes after start: \n$pid_new\n");
   return ($pid_new);
}



sub my_os {
##########
  # public os, hn, rh, path
  my ($t,@t,$uname,$rel);
  print "\n";

  $haszfs="no";
  $nc="nc";                                    # netcat

  # host and os name
  $hn=`hostname`; chomp($hn);                  # X and Windows
  $fn = $^O;                                   # family name from perl
  $on = $fn;                                   # on =family name from per view, later with details
  $dn=$fn;                                     # family name from perl, solaris, mswin32, darwin later, Illumos, Debian etc

  #defaults
  $os  ="other";
  $rh  ="OTH:$hn";
$tpath="/xampp/web-gui/tmp";


  if ($fn=~/Win32/i) {

      $t=`zpool list 2>&1`;
      if (!($t=~/\"zpool\"/s)) { $haszfs="yes"; } # missing "zpool" or zfs installed

      my $ver=`ver`;                        # Microsoft Windows [Version 10.0.19045.4046]

      $t=$ver;
      $t=~s/.*Version +//is;                # 10.0.1..

      $rh="WIN:$hn";                        # W1N:hostname
      $dn="Windows";

      $os=$ver;
      $nc="$spath/tools/nc/win/nc64";
      $nc=~s/\//\\/g;                      # windows cmd

  } else {

      # set nc to 755
      `chmod -R 755 $spath/*`;            # mainly nc

      $uname=`uname -mrs`;                # os, kernel, cpu
      chomp($uname);

      @t=split(/ +/,$uname);              # split

      chomp($path);

      #######################
      if ($fn=~/bsd|dragon|true/i) {    # bsd options
            $t=`zpool list 2>&1`;
            if (!($t=~/\"not found\"/s)) { $haszfs="yes"; } # missing "zpool" or zfs installed

            $on="bsd";
            if (($t[0]=~/free/i) && (!($t[1]=~/bsd/i))) {
              $t[1]="$t[0] $t[1]";
            }
            $on.=",$t[1],$t[2]";                       # os kernel, cpu

            $rh="BSD:$hn";

            $os="$uname";
            $nc="$spath/tools/nc/bsd/nc";
            if ($os=~/Free/) { $dn="Free-BSD"; }
            if ($os=~/Open/) { $dn="Open-BSD"; }

            #sambaconfig
            $sc=&get_smb_conf;
      }

      ########################
      if ($fn=~/Darwin/i) {        # OSX

           $t=`zpool list 2>&1`;
           if (!($t=~/\"not found\"/s)) { $haszfs="yes"; } # missing "zpool" or zfs installed

            $rh="OSX:$hn";
            $on.=",OSX,$t[2],$t[3]";        # kernel, cpu   Darwin xxx

            $os="macos";
            $os.=`sw_vers -productVersion`;              # macOS 12.6
            $dn="macos";

            $nc="$spath/tools/nc/osx/nc";
       }

       ########################
       if ($fn=~/Linux/i) {         # Linux options

          $t=`zpool list 2>&1`;
           if (!($t=~/\"not found\"/s)) { $haszfs="yes"; } # missing "zpool" or zfs installed

           # Proxmox
           if ($t[1]=~/pve/i) {
                $t[1]=~s/pve/pve Proxmox/;
                $rh="PVE:$hn";
           }

           $on.=",$t[1],$t[2]";         # kernel, cpu
           $os=$uname;                  # os, kernel, cpu
           $dn="Linux";

           if ($uname=~/86/) {
              $nc="$spath/tools/nc/lin/nc";
           }
           if ($os=~/pve/i)     { $dn='Proxmox'; }
           if ($os=~/Debian/i)  { $dn='Debian'; }
           if ($os=~/Ubuntu/i)  { $dn='Ubuntu'; }

           #sambaconfig
           $sc=&get_smb_conf;
       }

       ########################
       if ($fn=~/Solaris/i) {    # Solaris and forks
              # uname -mrs ->  always  SunOS 5.11 i86pc

              $haszfs="yes";
              my $buf="131072";

              if (-f "/etc/release") {
                 $t=`cat /etc/release`;
                 $t=~s/\n.*//s;                    # first line only with OS
                 $t=~s/^ +//;                     # no leading spaces
                 $t=~s/ *\(.*//;                  # no (opt text) ex in OI (powered..

                 #Solaris                         Oracle Solaris 11.4 X 86
                 if ($t=~/Solaris/i) {
                   $on.=",$t";
                   $os="Solaris ".`uname -v`;    chomp($os);
                   $rh="SOL:$hn";
                   $dn="Solaris_11";
                 }

                 # OmniOS                          OmniOS v11 r151046z
                 if ($t=~/OmniOS/i) {
                   $on.=",$t";
                   $os="OmniOS ".`uname -v`;    chomp($os);
                   $rh="OMN:$hn";
                   $dn="Illumos";
                 }

                 #SmartOS                         SmartOS 20240404T141941Z x86_64
                 if ($t=~/SmartOS/i) {
                   $on.=",$t";
                   $os=`uname -v`;  chomp($os);
                   $rh="SMA:$hn";
                   $dn="Illumos";
                 }

                 #OI                             OpenIndiana Hipster 2023.10 (powered..
                 if ($t=~/OpenIndiana/i) {
                   $on.=",$t";
                   $os="OI ".`uname -v`;
                   $rh="OI:$hn";
                   $dn="Illumos";
                 }

                 if ($os eq "")          {        # other Illumos
                    $on.=",$t";
                    $os="$t";
                    $rh="ILL:$hn";
                    $dn="Illumos";
                 }

              } else {      # no /etc/release

                  $on.=",$uname";
                  $os=$uname;
                  $rh="OSO:$hn";                 # other Solaris ??
                  $fn="solaris";
              }

              if ($uname=~/86/) {
                  $nc="$spath/tools/nc/sol/nc -b $buf";
              }

              $t=`svcs | grep smb/server`;      # kernelbased smb
              if ($t=~/online/s)  {
                 $sc="smb/server";
              } else {
                 #sambaconfig
                 $sc=&get_smb_conf;
              }


       } # /sol

  }  #/else non windows


  $rh=~s/\n//gs;
  $os=~s/\n//gs;
  $rh=~s/\n//gs;

  # callback
  if ($comm=~/callback/i) { $on.=",callback"; }

  print "\nzfs $haszfs\n";                  # zfs installed
  print "dn: $dn\n";
  print "on: $dn\;$rh\;$os\n";                        #  lx,bsd,callback   family name
  print "fn: $fn\n";                        #  lx,bsd
  print "hn: $hn\n";                         # hostname
  print "os: $os\n";
  print "rh: $rh\n";
  print "hi: $dn\;$rh\;$os\n";
  print "nc: $nc\n";
  print "sc: $sc\n";                         # samba conf
  print "auth: ".substr($data{'myauth'},0,4)."..\n\n";
  print "timeout: $timeout\n";
  print "max_connections: $max_connections\n";
  print "timeout: $timeout\n";
  print "allowed_ip: $allowed_ip\n";
  print "enforce_upsrv: $data{'enforce_upsrv'}\n";
  print "keyserver w1: $data{'w1'}\n";
  print "keyserver w2: $data{'w2'}\n";
  print "spath: $spath\n";
  print "path: $path\n";

  #perl INC
  print "\n\Perl \@INC\n";
  foreach my $dir (@INC) {
      print "$dir, ";
  }

  # ok
  print "\nzpool list\n".`zpool list`;
  print "\n\nwaiting for client requests..\n";

  return();

}


sub my_add {           # add ("path/filename","data",);  max 50 lines
##############

    my $file = $_[0];
    my $data = $_[1];
    my @old;

    ##########################
    my $afile = "$spath/tmp/activity.log";        # updated on file write
    my $lock="$spath/tmp/log.lock";
    open (X,"<$lock");
      flock(X, LOCK_EX);
          #
          open (F,"<$file");                      # old file
                flock(F, LOCK_EX);
                @old=<F>;
          close (F);

        my @new;
        for (my $i = 0; $i < @old; $i++) {
             #
             # Check if the maximum limit (100 elements) is reached
             if ($i >= 100) {
                 # Stop iterating if the limit is reached
                 last;
             }

             # Copy the current element from @old to @new_array
             push(@new, $old[$i]);
         }


         my $t=join("",@new);
         open (PF, ">$file");
            flock(PF, LOCK_EX);
            print PF "$t\n$data";                 # programm must add \n if needed to allow added strings
         close (PF);

         $t=time();
         open (PF, ">$afile");
            flock(PF, LOCK_EX);
            print PF "$t";                        # last  access
         close (PF);

    close (X);                                    # close lock


    #########################

    return();
}

sub my_append {           #   write ("path/filename","data",);
##############
    my $file = $_[0];
    my $data = $_[1];
    my $afile = "$spath/tmp/activity.log";         # updated on file write

    open(my $fh,">>$file");
      flock($fh, LOCK_EX);
        print $fh "$data\n";
      flock($fh, LOCK_UN);
    close($fh);

    $t=time();
    open(my $fh,">$afile");
      flock($fh, LOCK_EX);
        print $fh "$t";
      flock($fh, LOCK_UN);
    close($fh);

    return();
}

sub my_write {           #   write ("path/filename","data",);
##############
    my $file = $_[0];
    my $data = $_[1];
    my $afile = "$spath/tmp/activity.log";         # updated on file write

    open(my $fh,">$file");
      flock($fh, LOCK_EX);
        print $fh "$data\n";
      flock($fh, LOCK_UN);
    close($fh);

    my $t=time();
    open(my $fh,">$afile");
      flock($fh, LOCK_EX);
        print $fh "$t";
      flock($fh, LOCK_UN);
    close($fh);

    return();
}

sub my_read {           # read   read("path/filename");    opt return array
##############
    my $file = $_[0];
    my $arr  = $_[1];

    my $data="";
    my @data=();

    if (-f $file) {
        #
        open(my $fh, "<$file");
          flock($fh, LOCK_EX);
             @data = <$fh>;
          flock($fh, LOCK_UN);
        close($fh);


    } else {
         $data="$file not found";
    }

    if ($arr=~/a/) {  return(@data); }

    $data=join("",@data);
    return($data);
}

############   fileage tempfile in s
sub my_age {
###########
    # any write or add updates tmpfolder/activity.log with a date() stamp.
    # to sync with monitor.pl
    # this sub reads activity.log and returns age in s

    my $file = "$spath/tmp/activity.log";
    if (!(-f $file)) { return "-1"; }     # file not found

    my $r=&my_read($file);  # last access in s
    $t=time();
    my $old=$t-$r;  # current - last in s

    return($old);
}


###########   # send data via curl
sub my_updata {                          #
###########    # par ip, filename, auth, data      (-ks -> no cert check, silent)
   # curl -ks -X POST -F 'upfilefilename' -F 'data=abc' 'https://example.com/upload.cgi'

    my $upsrv=$_[0];
    my $filename=$_[1];         # filename without path
    my $auth=$_[2];
    my $text_data=$_[3];

    #$text_data= uri_escape($text_data);        # only for html uploads via Tiny etc

     print "\n".__LINE__." upload filename=$filename to $upsrv\n";
     print "curl --connect-timeout 5 -k -X POST -F \"upfile=$filename\" -F \"auth=$auth\" -F \"data=$text_data\" https://$upsrv/cgi-bin/cs/cs_connect.pl\n\n";

     my $r=`curl --connect-timeout 5 -k -X POST -F "upfile=$filename" -F "auth=$auth" -F "data=$text_data" https://$upsrv/cgi-bin/cs/cs_connect.pl`;
     print __LINE__." uploaded data to $filename on https://$upsrv/cgi-bin/cs/cs_connect.pl\n$r\n";
     if ($r=~/ok:/) { return ("$filename"); } else { return ("upload error: $r"); }

}

###########
###########   # send file via curl
sub my_upload {                          #
###########    # par ip, filename, auth, data      (-ks -> no cert check, silent)
   # curl -ks -X POST -F 'upfilefilename' -F 'data=@/path/to/your/file.txt' 'https://example.com/upload.cgi'

    my $up=$_[0];                    # ip
    my $filename=$_[1];              # filename without path
    my $auth=$_[2];
    my $ip=$_[3];

    my $k=" -k";                     # allow private cert
    if ($upsrv=~/[a-z]/) {           # from app setting, 'ip' or hostname
        if ($upsrv ne "ip") {        # self signed cert
          $up=$upsrv;
          $k="";
        }
    } #

    if ($up eq "") { $up=$ip; }
    #$text_data= uri_escape($text_data);        # only for html uploads via Tiny etc

    #enforce local settings
    if ($data{'_upsrv'} ne "")      { $up=$data{'_upsrv'}; }
    if ($data{'_upsrv_auth'} ne "") { $auth=$data{'_upsrv_auth'}; }

     print "\n".__LINE__." upload filename=$filename to $up via\n";
     print "curl --connect-timeout 10$k -X POST -F \"upfile=$filename\" -F \"auth=$auth\" -F \"data=\@$cpath/$filename\" https://$up/cgi-bin/cs/cs_connect.pl\n\n";

     my $r=`curl --connect-timeout 10$k -X POST -F "upfile=$filename" -F "auth=$auth" -F "data=\@$cpath/$filename" https://$up/cgi-bin/cs/cs_connect.pl`;
     print __LINE__." uploaded $filename to https://$up/cgi-bin/cs/cs_connect.pl\n$r\n";

     if ($r=~/ok:/) { return ("$filename"); } else { return ("upload error:"); }

}

###########   # get file content via curl
sub my_download {                          #
###########    # par ip, filename, auth, data      (-ks -> no cert check, silent)
    # curl -ks -X POST -F 'upfilefilename' -F 'data=@/path/to/your/file.txt' 'https://example.com/upload.cgi'
    # cmd for callback  rcmd::upsrv::filename::memberip::auth:::nocache::

    my $up=$_[0];
    my $filename=$_[1];              # filename without path
    $filename=~s/^rcmd:*//;          #rcmd: is only a upload marker, not part of filename
    my $auth=$_[2];
    my $ip=$_[3];

    my $k=" -k";                     # allow private cert
    if ($upsrv=~/[a-z]/) {           # from app setting, 'ip' or hostname
        if ($upsrv ne "ip") {        # self signed cert
          $up=$upsrv;
          $k="";
        }
    } #

    if ($up eq "") { $up=$ip; }      # web-gui Apache

    #enforce local settings
    if ($data{'_upsrv'} ne "")      { $up=$data{'_upsrv'}; }
    if ($data{'_upsrv_auth'} ne "") { $auth=$data{'_upsrv_auth'}; }

    #$text_data= uri_escape($text_data);        # only for html uploads via Tiny etc

    my $r=`curl --connect-timeout 10$k -X POST -F "getfile=$filename" -F "auth=$auth" https://$up/cgi-bin/cs/cs_connect.pl`;
    chomp($r);

    $t=substr($r,0.70);
    print "\n".__LINE__." download rcmdfile=$filename from $up via\n";
    print "curl --connect-timeout 10$k -X POST -F \"getfile=$filename\" -F \"auth=$auth\" https://$up/cgi-bin/cs/cs_connect.pl\nresult=$t\n";
    $r=~s/<n>/\n/gs;  # multiline transfer?


    if (($r eq  "") || ($r=~/error:/s))  {  return ("my_download error: or no data<hr>$r"); } else { return ($r); }

}

####################
sub c_filename {            # cache or command filename without ip and extension .data or .rcmd
###################
   my $cmd=$_[0];           # zpool list ..          #
   $cmd=~s/^ +| +$//g;      # unneded spaces
   $cmd=~s/ *line:=.*//;

   my $mem=$_[1];           #member 192...          #

  # get filename ip_cmd_hash        # max 64 B  ex: 127.0.0.1_zfs_get_all_623924ecd6b9239f336cc73d7691a219f.data

  my $h  = sha256_hex($cmd);        # sha of cmd
  $t=substr($cmd,0,20);             # first 30 char of cmd
  my $cfn="$mem\_$t\_$h";            # (ip)_cmd_hash
  $cfn=substr($cfn,0,55); # shorten to 55, real cmd is verified via file.last
  #&mess($cfn);
  $cfn=~s/ +/_/gs;
  $cfn=~s/[^a-zA-Z0-9\.\_\-]//gs;
  return($cfn);
}

###########   # send large data via post and Tiny
sub my_post {                          #
###########    # par ip, filename,data

    my $ip=$_[0];
    my $filename=$_[1];
    my $text_data=$_[2];         # plain text or data must be encoded.
    $text_data= uri_escape($text_data);

    my $len=length($text_data);

   # if ($filename=~/\.stat/) {
   #     $filename=~s/\.stat.*/.stat/;
   # }

    #debug
    #$filename="abc.tmp";    #   -> my_ip~abc.tmp

    # Data escaped
    # use URI::Escape;

    # $data_decoded = uri_unescape($data_encoded);

    # use HTTP::Tiny;  # 0.08    # in mainscript

    # Define the URL of the Apache script
    my $url = "http://$ip/cgi-bin/admin.pl";

    print "654 sub my_post, must upload $len Bytes data to $filename\n";

    my $boundary = "Boundary_".time();  # boundary string

    # HTTP-Request
    my $http = HTTP::Tiny->new;
    my $response = $http->request('POST', $url, {
        headers => {
            'Content-Type' => "multipart/form-data; boundary=$boundary",
        },
        content => "--$boundary\r\n" .
                    "Content-Disposition: form-data; name=\"file\"; filename=\"fn.txt\"\r\n" .
                    "Content-Type: text/plain\r\n\r\n" .
                    "$text_data\r\n" .

                    "--$boundary\r\n" .    # add formelement
                    "Content-Disposition: form-data; name=\"action\"\r\n\r\n" .
                    "rgroup_data_$filename\r\n" .  # filename
                    "--$boundary--\r\n",

    });


    my $response_body = $response->{content};
    #$response_body   = uri_unescape($response_body);                   # reverse escape

    # process answer
    if ($response->{success}) {
        print __LINE__." data uploadad, $response_body\n";
        return ($filename);
    } else {
        print __LINE__." upload error: ", $response->{status} || "other error", " (599 is network connect timeout)\n$response_body\n";
        print `ping $ip`."\n";
        return ("upload error: $response->{status}");
    }

}

#####################
sub client_ip_allowed    {
####################
my $client_ip=$_[0];

        print "\n".__LINE__." check if $client_ip is allowed: ";
        if ($client_ip eq "127.0.0.1") {
               print "always\n";
               return (1);
        }

        if ($allowed_ip ne "") {
           $t=$allowed_ip ;
           $t=~s/\./\\./g;                    # respect .
           $t=~s/ *, */|/g;                   # allow commalist ex 192.2.1.1,172.16, (this must be part of ip)
           if (!($client_ip=~/$t/)) {
             print "no\n";
             return (0);                      # not allowed
           }
        }
        print "yes\n";
        return (1);
}                                             # allowed

sub get_mydate {                              # setzt datumsvariablen
###########
 my ($sec,$min,$hour,$mday,$mon,$year,$wday,$ydat,$isdst) = localtime(time);
 $mon++;
 if ($sec < 10) { $sec = "0$sec"; }
 if ($min < 10) { $min = "0$min"; }
 if ($hour < 10) { $hour = "0$hour"; }
 if ($mday < 10) { $mday = "0$mday"; }
 if ($mon < 10) { $mon = "0$mon"; }
 $year=1900+$year;
 return("$year\.$mon\.$mday\.$hour\.$min\.$sec");
}

#
1;