#!
use strict;         # check for proper code

###############
sub my_action {
###############     build actions according to this template

  my ($out,$var,@p,$k,$u);  # declare all your variables as local or you will get errors

  # menue actions
  ###############
  &load_lib('zfslib.pl');

  ##############################
  # info
  ##############################

my $m=<<EoF;
<b>Dedup</b>
Realtime dedup is one of the killer features of ZFS as it can massively reduce required capacity for redundant datablocks.
This also can increase performance as there is less data that must be written or read from pool.
While dedup can be enabled per filesystem, it creates dedup tables that work poolwide. Once enabled, you cannot disable.

<b>Traditional Dedup</b>
Beside dedup2 in native Oracle Solaris ZFS, dedup was in the past a feature that should be avoided in most cases as it
can eat up all RAM with a catastrophic impact on performance and even with smaller dedup tables it affected performance negatively
without options to limit or clean up dedup tables. A special dedup vdev can limit the negative impact on RAM usage and performance..

<b>Fast Dedup in Open-ZFS</b>
The new Fast Dedup feature can be a gamechanger as it may allow to enable dedup in most cases just like compress with mostly more advantages
than disadvantages as RAM usage, DDT sizes, DDT cleanup and overall performance has seen massive improvements.
You can enable Fast Dedup in menu Pools > ZFS > Features (newest Open-ZFS)

<b>Fast dedup pool properties</b>
dedup_table_size:  the total size of all DDTs on the pool and
dedup_table_quota: the maximum possible size of all DDTs in the pool

When set, quota will be enforced by checking when a new entry is about to be created. If the pool is over its dedup quota, the entry won't be created,
and the corresponding write will be converted to a regular non-dedup write. Note that existing entries can be updated (ie their refcounts changed),
as that reuses the space rather than requiring more.

dedup_table_quota can be set to auto, which will set it based on the size of the devices backing the "dedup" allocation class. This makes it possible to limit
the DDTs to the size of a dedup vdev only, such that when the device fills, no new blocks are deduplicated.

Fast Dedup is currently in Open-ZFS 2.2.6 master and available soon (already in Open-ZFS on Windows 2.2.6 rc)
<b>Without Fast Dedup (Open-ZFS) or Dedup2 (Solaris), dedup should be avoided in most cases</b>.


EoF

#url
my $url="<a title=\"=title=\" class=\"=class=\" href=\"/cgi-bin/admin.pl?id=$in{'id'}\&amp;action=\&amp;member=$in{'member'}&amp;l1=$in{'l1'}\&amp;l2=$in{'l2'}\&amp;l3=$in{'l3'}\&amp;list==list=\&amp;zfs==zfs=\&amp;prop==prop=\&amp;prop2==prop2=\&amp;old==old=\">=text=</a>";


###########
print <<EoF;

  About Dedup
  <a href="#" onclick="\$('.info').toggle();"><img src="/_doc/menu/ico/i16.png"></a>
  <div class='info'><pre>$m</pre></div>

EoF
##########

if ($current{'on'}=~/Solaris/i) {
  print "Fast Dedup is not available on Solaris.<br>Just enable Dedup (dedup2)";
  &log_end;
  exit;
}

# get fast dedup
my $fd=&exe("zpool get -H -o name,property,value feature\@fast_dedup,dedup_table_size,dedup_table_quota,dedupcached");
#tank    feature@fast_dedup      enabled
#tank    dedup_table_size        0
#tank    dedup_table_quota       auto

@t=split(/\n/,$fd);
foreach my $t (@t) {
  if (!($t=~/\w/)) { next; }
  @p=split(/\s+/,$t);

  $k="$p[0]\_$p[1]";   # fd: enabled, disabled or ""  (n.a.)
  $zfs{$k}=$p[2];
}

#&print_hash(%zfs);
my $zfs=&exe("zfs list -H -o name,dedup");
my @zfs=split(/\n/,$zfs);
foreach my $z (@zfs) {
    @p=split(/\s+/,$z);
    $k="$p[0]\_dedup";
    $zfs{$k}=$p[1];
}

#change zfs property
if ($in{'answered'})           { &zfslib_change_property; }

#table
$out="Filesystem\tFastDedup Feature\tDedup\tDedupratio\tDDT_Size\tDDT_Quota\tDedupcached in Arc\n";

foreach my $z (@zfs) {

   @p=split(/\s+/,$z);

   # fs
   $out.="$p[0]\t";

   # feature
   if ($p[0]=~/\//) {
         $out.=" \t";
   } else {
         $k="$p[0]\_feature\@fast_dedup";
         if ($zfs{$k} eq "") {
            $out.="n.a.\t";   # feature not avail
         } else {
            if ($zfs{$k} eq "enabled") {
                $out.="<b>$zfs{$k}</b>\t";
            } else {
               $out.="$zfs{$k}\t";
            }
         }
   }

   # dedup state
   #############
   $k="$p[0]\_dedup";

   $u=$url;
   $u=~s/=zfs=/$p[0]/;
   $u=~s/=prop=|=title=/dedup/g;

   if ($zfs{$k} eq "on") {
      $t="<b>$zfs{$k}</b>";
   } else {
      $t="$zfs{$k}";
   }

   $u=~s/=text=/$t/;
   $u=~s/=old=/$zfs{$k}/;

   $out.="$u\t";


   # dedup ratio
   #############
   $k="$p[0]\_dedupratio";
   $out.="$zfs{$k}\t";

   # ddtsize
   $k="$p[0]\_dedup_table_size";
   if ($p[0]=~/\//) {
      $out.=" \t";
   } else {
      $out.="<b>$zfs{$k}</b>\t";
   }

   # ddtquota
   $k="$p[0]\_dedup_table_quota";

   $u=$url;
   $u=~s/=zfs=/$p[0]/;
   $u=~s/=prop=|=title=/ddtquota/g;
   $u=~s/=text=/$zfs{$k}/g;
   $u=~s/=old=/$zfs{$k}/g;

   if ($p[0]=~/\//) {
      $out.=" \t";
   } else {
      $out.="$u";
   }
   if ($zfs{$k} eq "auto") {
      $out.=" (max,dedup_vdev_size)";
   }
   $out.="\t";

   #cached
   $k="$p[0]\_dedupcached";
   if ($p[0]=~/\//) {
      $out.=" \t";
   } else {
      $out.="<b>$zfs{$k}</b>\t";
   }

   $out.="\n";
}

print "<br><br><b>Fast Dedup</b>";
print &list2table($out);
print "<br>You can enable Fast Dedup in menu Pools > ZFS > Features (newest Open-ZFS)";

# info
print "<script language='javascript'>document.getElementById('hl').innerHTML='Fast Dedup'</script>\n";
&log_end;

#ask property
if (($in{'prop'} ne "") && ($in{'answered'} ne "1")) {
    print &zfslib_ask_property;
    print "\n</body></html>";
    exit;
}

if ($log_on) {  &print_hash(%zfs); }
} #/ my_action




###############################################
#end of my script, do not delete the line below
1;
