 # get-disk.pl
#
# (c) napp-it 2010-24
# windows feb 2024

 use strict;
 my ($diff,$last,$disk2, $on,@p);

# os family name to get disks
$on=&exe('on');
$disk{'on'}=$on;

$disk{'all_infos_from'}=$in{'member'};

#Windows family
if ($on=~/Win/i) {

   # properties for disk overview (require $disk{'disk_property'})
   $disk{'header'}="DeviceID,UniqueId,BusType,Usage,Vendor,Model,MediaType,CanPool,CannotPoolReason,ZFS,Status,HealthStatus,Serialnumber,Size,hide";
   &get_win_disk;                            # add disk{'all'} from windows
}

if ($on=~/Solaris/i) {     # tools format, iostat -Enr, diskinfo
    &get_sol_disk;
    $disk{'header'}="DeviceID,Vendor,Model,Product,Revision,Serial_No,Hard_Errors,Soft_Errors,Transport_Errors,Size,Poolmember";
}

if ($on=~/Linux/i) {     # tools lsblk, smartctl
    &get_lx_disk;
    $disk{'header'}="DeviceID,Vendor,Model,Product,Revision,Serial_No,Size,Poolmember";
}

if ($on=~/bsd/i) {     # tools camcontrol devlist, smartctl
    &get_bsd_disk;
    $disk{'header'}="DeviceID,Vendor,Model,Product,Revision,Serial_No,Size,Poolmember";
}

if ($on=~/darwin/i) {     # disklist, smartctl
    &get_darwin_disk;
    $disk{'header'}="DeviceID,Vendor,Model,Product,Revision,Serial_No,Size,Poolmember";
}


if ($disk{'on'} eq "") {
    $disk{'on'}="unknown";
}

#&get_used_avail;  # -> get_zpool

#all_unused=all - all_poolmember
$disk{'all_unused'}=$disk{'all'};
@t=split(/\t/,$disk{'all_poolmember'});
foreach my $t (@t) {
    $disk{'all_unused'}=~s/$t\t*//;
}

# remove from unused when !canpool
if ($on=~/Win/i) {          # unused must mean available for windows pool
   @p=split(/\t/,$disk{'all_unused'});
   foreach my $p (@p) {
      if (!($disk{'all_canpool'}=~/$p/)) {
        $disk{'all_unused'}=~s/$p\t*//;
      }
   }
}

$disk{'all_unused'}=~s/\t+$|^\t+//;

$disk{'all_removable'}=~s/mirror\t+//;
$disk{'all_removable'}=~s/^t+//;

#&print_hash(%disk);

 #################
sub get_darwin_disk {
#################

    # step 1: lsblk | grep disk        (get all disks)
    # step 2: smartctl -i /dev/sda     (datails)


      my (@w,$w,$key,$k,$n,@n,@p,$p,$r,$i,$v,@v,$d,@d,$id,$out,$pool,$wd);
      $disk{'all'}="";
      $disk{'all_unused'}="";       #

      $r=&exe("diskutil list");     # liefert diskliste disk0 disk1

      #/dev/disk0 (internal, physical):
      #:                       TYPE NAME                    SIZE       IDENTIFIER
      #0:                                                   *2.1 GB     disk0

      #/dev/disk1 (internal, physical):
      #:                       TYPE NAME                    SIZE       IDENTIFIER
      #0:      GUID_partition_scheme                        *85.9 GB    disk1
      #1:                        EFI ⁨EFI⁩                     209.7 MB   disk1s1
      #2:                 Apple_APFS ⁨Container disk3⁩         85.7 GB    disk1s2

      # search physical
      @t=split(/\n/,$r);
      foreach my $t (@t) {
          if ($t=~/physical/)  {
              $t=~s/ .*//;
              $t=~s/\/dev\///;     # disk0
              $disk{'all'}.="$t\t";
          }
      }
      $disk{'all'}=~s/\t+$//;

#&mess($disk{'all'});       # debug


      @d=split(/\t/,$disk{'all'});
      &get_smart_from_all(@d);      # read smart for disks
}

#################
sub get_bsd_disk {
#################

    # step 1: lsblk | grep disk        (get all disks)
    # step 2: smartctl -i /dev/sda     (datails)


      my (@w,$w,$key,$k,$n,@n,@p,$p,$r,$i,$v,@v,$d,@d,$id,$out,$pool,$wd);
      $disk{'all'}="";
      $disk{'all_unused'}="";       #

      $r=&exe("sysctl -n kern.disks");     # liefert diskliste auf verschiedenen BSD
      #&mess($r);       # debug
      # ada0 ada1 freebsd, dragonfly, ghostbsd,trueos
      # wd0 wd1 openbsd
      # ws0 sd0 netbsd



      @t=split(/ +/,$r);
      foreach my $t (@t) {
           #print "$t<br>";
           if ($t=~/\w/) {
              $t=~s/.*,//;
              $t=~s/\)//;
              push (@d,$t);
           }
      }
      #&print_array(@d);

      $disk{'all'}=join("\t",@d);
      &get_smart_from_all(@d);      # read smart for disks
}

#################
sub get_lx_disk {
#################

    # step 1: lsblk | grep disk        (get all disks)
    # step 2: smartctl -i /dev/sda     (datails)


      my (@w,$w,$key,$k,$n,@n,@p,$p,$r,$i,$v,@v,$d,@d,$id,$out,$pool,$wd);
      $disk{'all'}="";
      $disk{'all_unused'}="";       #

      $r=&exe("lsblk");
      @t=split(/\n/,$r);
      foreach my $t (@t) {
           if ($t=~/^sd/) {
              $t=~s/ .*//;
              push (@d,$t);
           }
      }

      $disk{'all'}=join("\t",@d);
      &get_smart_from_all(@d);      # read smart for disks
}

#######################
sub get_smart_from_all {      # bsd, linux
######################

      my @d=@_;
      #
      # get details via smartctl
      my ($r,$w,$k,@w);

      foreach my $d (@d) {
             if ($on=~/darwin/i) {
                $r=&exe("/usr/local/sbin/smartctl -A /dev/$d");
             } else {
                $r=&exe("smartctl -i /dev/$d");
             }

             #Vendor:               VMware
             #Product:              Virtual disk
             #Revision:             2.0
             #Compliance:           SPC-4
             #User Capacity:        16,106,127,360 bytes [16.1 GB]
             #Logical block size:   512 bytes
             #LU is fully provisioned
             #Rotation Rate:        15000 rpm
             #Device type:          disk
             #Local Time is:        Mon Feb 26 15:37:56 2024 CET
             #SMART support is:     Unavailable - device lacks SMART capability

             #Model Family: Seagate Barracuda
             #Device Model: ST2000DM008-2FR102
             #Serial Number: 5VC34298
             #LU WWN Device Id: 5 0014ee 000000000
             #Firmware Version: 3.00
             #User Capacity: 2,000,418,886,400 bytes [2.00 TB]
             #Sector Size: 512 bytes logical/physical
             #Device is: In smartctl database [for details use: -P showall]
             #ATA Version is: 8
             #ATA Standard is: ATA-8-ACS revision 4
             #Local Time is: Sat Feb 25 15:56:08 2024 CET
             #SMART support is: Available - device has SMART capability.
             #SMART support is: Enabled

             @w=split(/\n/,$r);

             foreach my $w (@w) {   # lines
                 @p=split(/:/,$w);  # parameter xx: value

                 if ($p[0]=~/\bVendor\b|\bProduct\b|\bRevision\b|\bCompliance\b/i) {
                      $k=$d."_$p[0]";
                      $disk{$k}=$p[1];
                 }

                 if ($p[0]=~/Serial/i) {
                      $k=$d."_Serial_No";
                      $disk{$k}=$p[1];
                 }

                 if ($p[0]=~/Model/i) {
                      $k=$d."_Model";
                      $disk{$k}=$p[1];
                 }

                 if ($p[0]=~/Device/i) {
                      $k=$d."_Device";
                      $disk{$k}=$p[1];
                 }

                 if ($p[0]=~/ATA Standard/i) {
                      $k=$d."_ATA";
                      $disk{$k}=$p[1];
                 }

                 if ($p[0]=~/Firmware/i) {
                      $k=$d."_Firmware";
                      $disk{$k}=$p[1];
                 }

                 if ($p[0]=~/WWN/i) {
                      $k=$d."_WWN";
                      $disk{$k}=$p[1];
                 }

                 if ($p[0]=~/Sector/i) {
                      $k=$d."_Sectorsize";
                      $disk{$k}=$p[1];
                 }

                 if ($p[0]=~/Smart support/i)   {
                      $k=$d."_smart";
                      if ($p[1]="unavailable") {
                         $disk{$k}="unavailable";
                      } else {
                         if ($disk{$k} =~/Enabled/i) {
                            $disk{$k}="enabled";
                         }

                         $disk{$k}="available";
                      }
                 }

                 if ($p[0]=~/Capacity/i)  {
                      $k=$d."_Size";

                      if ($p[1]=~/\[\d/) {        #16,106,127,360 bytes [16.1 GB]
                         $p[1]=~s/.*\[//;
                         $p[1]=~s/\]//;           #16.1 GB
                      }
                      $disk{$k}=$p[1];
                }

             }
             #&mess($r);
      }


      #$disk{'all_unused'}=$disk{'all_unused'};
      #&print_array(@t);
      #&print_hash(%disk);
      #&mess($r);
}

sub get_used_avail_xx {
###################

    my $pool=&exe("zpool status");
    my @all=split(/\t/,$disk{'all'});          # all disks

    $disk{'all_unused'}=$disk{'all'};            # start with, now remove pool members

    foreach my $d (@all) {
            if ($pool=~/$d/s)   {
               $disk{'all_unused'}=~s/$d\t*//;
            }
    }
}

##################      # solaris
sub get_sol_disk {
#################

  my (@w,$w,$key,$k,$n,@n,@p,$p,$r,$i,$v,@v,$d,@d,$id,$out,$pool,$wd);
  $disk{'all'}="";
  $disk{'all_unused'}="";       #

  $r=&exe("format");
  @t=split(/\n/,$r);

  #[4] -> '       0. c1t0d0 <VMware-Virtual disk-2.0-35.00GB>'
  #[5] -> '          /pci@0,0/pci15ad,1976@10/sd@0,0'
  #[6] -> '       1. c3t1d0 <INTEL-SSDPEDME400G4-8DV10171-372.61GB>'
  #[7] -> '          /pci@0,0/pci15ad,7a0@16/pci8086,3709@0/blkdev@1,0'

  $i=0;
  foreach my $t (@t) {
      if (!($t=~/^\s+\d+\. /))   { next; }
      $t=~s/^\s+//;
      @p=split(/ +/,$t);
      $disk{'all'}.="$p[1]\t";
  }                                              # all current disks

  $disk{'all'}=~s/\t+$//;



  # iostat -Enr, known disks since boot
  $r=&exe("iostat -Enr");

  #c3t1d0           ,Soft Errors: 0 ,Hard Errors: 4 ,Transport Errors: 0
  #Model: INTEL SSDPEDME400G4                      ,Revision: 8DV10171 ,Serial No: CVMD5450001D400AGN
  #Size: 400.09GB <400088457216 bytes>
  #Media Error: 0 ,Device Not Ready: 0 ,No Device: 0 ,Recoverable: 0
  #Illegal Request: 0 ,Predictive Failure Analysis: 0


  @t=split(/\n/,$r);
  my (@a,@b,@c,@d,@e);      # prop per disk
  my (@a1,@a2,@a3,@a4);     # prop:value
  my (@b1,@b2,@b3,@b4);     # prop:value
  my (@c1,@c2,@c3,@c4);     # prop:value
  my (@d1,@d2,@d3,@d4);     # prop:value
  my (@e1,@e2,@e3,@e4);     # prop:value

  $i=0;
  foreach my $t (@t) {

     # process disk
     @a=split(/,/,$t[$i]);          # diskid , hard: n , soft: n , transport : n
     $d=$a[0];                      # this disk
     $d=~s/ +//g;
     @b=split(/,/,$t[$i+1]);        # size: xx
     @c=split(/,/,$t[$i+2]);        # media error: n, pred.. : n, ..
     @d=split(/,/,$t[$i+3]);

     @a1=split(/:/,$a[0]);          # disk
     @a2=split(/:/,$a[1]);          # soft error: n
     @a3=split(/:/,$a[2]);          # hard error: n
     @a4=split(/:/,$a[3]);          # trans err: n

     $a2[0]=~s/ /_/g;               # soft_error
     $a3[0]=~s/ /_/g;               # hard_error
     $a4[0]=~s/ /_/g;               # transport_error

     $k=$d."_Soft_Errors";
     if ($d ne "") { $disk{$k}=$a2[1]; $disk{$k}=~s/ //g;  }

     $k=$d."_Hard_Errors";
     if ($d ne "") { $disk{$k}=$a3[1]; $disk{$k}=~s/ //g;  }

     $k=$d."_Transport_Errors";
     if ($d ne "") { $disk{$k}=$a4[1]; $disk{$k}=~s/ //g;  }

     $k=$d."_Soft_Errors";
     if ($d ne "") { $disk{$k}=$a2[1]; $disk{$k}=~s/ //g;  }

     # $k=Model, Product, Vendor, Revision, Serial are option s on line 2 (b)
     @b1=split(/:/,$b[0]);          # Produkt : xx         (2-4 values per line)
     @b2=split(/:/,$b[1]);          # Revision: xx
     @b3=split(/:/,$b[2]);          # Vendor: n
     @b4=split(/:/,$b[3]);          # Serial No: n

     if ($b1[0] ne "") { $b1[0]=~s/ /_/g; $k=$d."_$b1[0]"; $disk{$k}=$b1[1] }    # Vendor or Serial No
     if ($b2[0] ne "") { $b2[0]=~s/ /_/g; $k=$d."_$b2[0]"; $disk{$k}=$b2[1] }    # Vendor or Serial No
     if ($b3[0] ne "") { $b3[0]=~s/ /_/g; $k=$d."_$b3[0]"; $disk{$k}=$b3[1] }    # Vendor or Serial No
     if ($b4[0] ne "") { $b4[0]=~s/ /_/g; $k=$d."_$b4[0]"; $disk{$k}=$b4[1] }    # Vendor or Serial No

     @c1=split(/:/,$c[0]);          # Size: xx
     @c2=split(/:/,$c[1]);          #
     @c3=split(/:/,$c[2]);          #
     @c4=split(/:/,$c[3]);          #

     if ($c1[0] ne "") { $c1[0]=~s/ /_/g; $k=$d."_$c1[0]"; $disk{$k}=$c1[1] }    # Size
     if ($c2[0] ne "") { $c2[0]=~s/ /_/g; $k=$d."_$c2[0]"; $disk{$k}=$c2[1] }    #
     if ($c3[0] ne "") { $c3[0]=~s/ /_/g; $k=$d."_$c3[0]"; $disk{$k}=$c3[1] }    #
     if ($c4[0] ne "") { $c4[0]=~s/ /_/g; $k=$d."_$c4[0]"; $disk{$k}=$c4[1] }    #

     if ($d ne "") {
       $k=$d."_Size"; $disk{$k}=~s/^ +//; $disk{$k}=~s/ .*//;   # ' 8001.56GB <8001563222016 bytes>'
     }

     @d1=split(/:/,$d[0]);          # Media Error: xx
     @d2=split(/:/,$d[1]);          #
     @d3=split(/:/,$d[2]);          #
     @d4=split(/:/,$d[3]);          #

     if ($d1[0] ne "") { $d1[0]=~s/ /_/g; $k=$d."_$d1[0]"; $disk{$k}=$d1[1] }    # Media Error
     if ($d2[0] ne "") { $d2[0]=~s/ /_/g; $k=$d."_$d2[0]"; $disk{$k}=$d2[1] }    #
     if ($d3[0] ne "") { $d3[0]=~s/ /_/g; $k=$d."_$d3[0]"; $disk{$k}=$d3[1] }    #
     if ($d4[0] ne "") { $d4[0]=~s/ /_/g; $k=$d."_$d4[0]"; $disk{$k}=$d4[1] }    #

     @e1=split(/:/,$e[0]);          # Illegal Request: xx
     @e2=split(/:/,$e[1]);          #
     @e3=split(/:/,$e[2]);          #
     @e4=split(/:/,$e[3]);          #

     if ($e1[0] ne "") { $e1[0]=~s/ /_/g; $k=$d."_$e1[0]"; $disk{$k}=$e1[1] }    # Illegal Request
     if ($e2[0] ne "") { $e2[0]=~s/ /_/g; $k=$d."_$e2[0]"; $disk{$k}=$e1[1] }    #
     if ($e3[0] ne "") { $e3[0]=~s/ /_/g; $k=$d."_$e3[0]"; $disk{$k}=$e1[1] }    #
     if ($e4[0] ne "") { $e4[0]=~s/ /_/g; $k=$d."_$e4[0]"; $disk{$k}=$e1[1] }    #


     $i=$i+5;
  }


  #&mess($r);
  #&print_hash(%disk);

}


##############      PHYSICALDRIVE1
sub get_win_disk  {
##############
  my (@w,$w,$key,$k,$n,@n,@p,$p,$r,$i,$v,@v,$d,@d,$id,$out,$pool,$wd);
  $disk{'all'}="";
  $disk{'all_unused'}="";       # 0 1 2

  ##########################################################
  # Diskinfo via powershell Get-PhysicalDisk | ConvertTo-csv (all)
  ##########################################################

  #wanted data

  # read all data
  $r=&exe("Get-PhysicalDisk | ConvertTo-csv");

  # get headers
  $disk{'disk_headers'}=&csv2par($r);                                 # FriendlyName or UniqueId as key  (first value)

  $w='DeviceId,UniqueId,Manufacturer,Model,MediaType,CanPool,OperationalStatus,HealthStatus,Serialnumber,Size,Usage';

  # 9,600144F06DAD1770000066768F3E0009,SUN,COMSTAR,HDD,False,OK,Healthy,,0,Auto-Select
  # 5,600144F06DAD17700000666EB8940003,SUN,COMSTAR,SSD,False,OK,Healthy,,0,Auto-Select

  #get disk data for $w
  my $dd=&csv2data($r,$w,'physicaldrive:DeviceId');            # csv, headers,pre:key for %win ex p:poolname_value

  #Windows disk
  $wd=&exe("ps (Get-Partition -DriveLetter (Get-Item (Get-WmiObject Win32_OperatingSystem).SystemDrive).PSDrive.Name).DiskNumber");
  # &mess($wd);

  $wd=~s/\D//gs;         # windows disk is for ex 0
  $disk{'osdisk'}="physicaldrive$wd";
  $disk{'hide'}=$cfg{'value_hide_disk'};       # hide disk  ex physicaldrive3,physicaldrive4
  #$disk{'hide'}.=",physicaldrive$wd";
  #$disk{'hide'}=~s/^,//;

  $t="physicaldrive$wd\_osdisk";
  $disk{$t}="1";                           # Win bootdisk


  @w=();
  @w=split(/\n/,$dd);     # diskdata
  @w=sort(@w);

  # &print_array(@w);
  # [0] -> '0,eui.0025388C91D12047,,SAMSUNG MZVLB512HBJQ-000L7,SSD,False,OK,Healthy,,0,Auto-Select'
  # [1] -> '1,600144F06DAD17700000666FED8F0004,SUN,COMSTAR,SSD,False,Transient Error OK,Unhealthy,,0,Auto-Select'
  # [2] -> '2,600144F06DAD17700000666FED900005,SUN,COMSTAR,HDD,False,OK,Healthy,,0,Hot Spare'

  foreach $w (@w) {     # each disks
     if (!($w=~/^\d/))    { next; }

     @p=split(/\,/,$w);  # parameter

     # &print_array(@p);
     # [0] -> '0'
     # [1] -> 'eui.0025388C91D12047'
     # [2] -> ''
     # [3] -> 'SAMSUNG MZVLB512HBJQ-000L7'
     # [4] -> 'SSD'
     # [5] -> 'False'
     # [6] -> 'OK'
     # [7] -> 'Healthy'
     # [8] -> ''
     # [9] -> '0'
     # [10] -> 'Auto-Select'

     foreach $p (@p) {           # each parameter

          $id="physicaldrive$p[0]";             # ex physicaldrive0

          #all
          if (!($disk{'all'}=~/$id\b/)) {                                 # all disks, nvme are missing
             $disk{'all'}.="$id\t";
          }

          # canpool = true, -> available for pooling (win and zfs)
          if ($p[5] eq "True")  {
            if (!($disk{'all_canpool'}=~/$id\b/i)) {
              $disk{'all_canpool'}.="$id\t";
            }
          }

          $key="$id\_$n[0]";
          $disk{$key}=$n[1];       # 0_size -> 1234

          if (!($disk{'hide'}=~/$id\b/i)) {                         # hide disk?
               # print "not hide: $id in $disk{'hide'}<br>";

               if (!($disk{'all_unused'}=~/$id\b/)) {               # not yet included
                  if ($id ne "$disk{'osdisk'}") {                   # ex physicaldrive0 = Windows bootdisk
                     if (!($disk{'all_poolmember'}=~/$id\b/s)) {    # not a zfs pool member

                        $disk{'all_unused'}.="$id\t";
                        # print "-$id- -> -$disk{'osdisk'}-<br>";
                     }
                  }
               } #/unused

        #       #print "hide: $id in $disk{'hide'}<br>";
        #       $key="$id\_hide";                                   # hide
        #       $disk{$key}="1";


               # disk_Size like on all other platforms for pool overview
               $key="$id\_Size";
               #$t="$id\_dcap";
               $disk{$key}=$disk{$key};

          } #/hide
     } #/eack parameter
  }  #/each disk

  $disk{'all'}=~s/\t+$//;
  $disk{'all_unused'}=~s/\t+$//;
  $disk{'all_poolmember'}=~s/\t+$//;
  $disk{'all_canpool'}=~s/\t+$//;

  # now we must remove all disks from unused  that are not in all_canpool
  # as unused  must mean not in a zfs pool and available for a Windows pool

  @t=split(/\t/,$disk{'all_unused'});        # only from zfs view
  @w=();
  foreach my $t (@t) {
      if ($disk{'all_canpool'}=~/$t/) { push (@w,$t); }
  }
  $disk{'all_unused'}=join("\t",@w);


  # all_winremoveable (shrink pool) = all - zfs poolmember - unused - osdisk
  $disk{'all_winremovable'}=$disk{'all'};

  # zfs poolmember
  @t=split(/\t/,$disk{'all_poolmember'});
  foreach my $t (@t) {
      $disk{'all_winremovable'}=~s/$t\t*//;
  }

  # unused
  @t=split(/\t/,$disk{'all_unused'});
  foreach my $t (@t) {
      $disk{'all_winremovable'}=~s/$t\t*//;
  }

  # osdisk
  $disk{'all_winremovable'}=~s/$disk{'osdisk'}\t*//;


  #&print_hash(%disk);
  #&print_array(@w);
  #&mess($r);

}


#####################          !!
sub get_mydisk_bcap { #
###################
  my $id=$_[0];                           # diskid
  my $bytes=$_[1];                        # bytes
  my ($ib,$iu,$b,$u,$k);

 # &mess($bytes);
# &print_hash(%disk);
  my $dez=&b_convert("$bytes","b","s");
  my $bin=&b_convert("$bytes","b","is");

  $k="$id\_dcap";
  $disk{$k}="$dez";
  if (!($dez =~/\d/)) { $disk{$k}="-"; }


  $k="$id\_bcap";
  $disk{$k}="$bin";
  if (!($bin =~/\d/)) { $disk{$k}="-"; }

}

# eof
1;