#!
# (c) guenther alka 2017-2024
#

# job for reports

# debug: run from console
# perl /var/web-gui/data/menues/_lib/window/scripts/job-report.pl run_123 456
# 123= jobid, 456=time()


use strict;
use Fcntl qw(:flock);                                          # filelocking
$|=1;                                                          # keine Pufferung, no buffering

use vars qw($wpath $dpath $tpath $par $id $jobid $opt $action $os $f $t @t $p @p @reports $reports $alertmess @run %sys %cfg %current $error $alert $alertlist $reasons $info $message $text $subject $smtp $ref $path);

$wpath="/xampp/web-gui";
$dpath="/xampp/web-gui/data";
$tpath="/xampp/web-gui/tmp";

use lib "/xampp/web-gui/data/cs_server/CGI";
use Digest::SHA::PurePerl qw(sha256_hex);                    # replaces md5m, $hash = sha256_hex($data); 96 char

require "$dpath/menues/_lib/windows/task-lib.pl";            #
require "$wpath/data/wwwroot/cgi-bin/admin-wlib.pl";
require "$dpath/menues/_lib/windows/zfslib.pl";

$par=$ARGV[0];
 if (!(defined($par))) {
     &my_log("job-report:  missing parameter");
      exit;
 }

 # get cfg (auth + cfg settings like mail or upsrv) from adminlib per single report
 # &get_cfg; # with $current{'member_ip'}="";      # per member call


 ############

 #par: action_job timestamp
 $par=~s/ .*//;

 # parameter: action_$job time
 if (!($par =~/\w+\_\d+$/)) {
     print "job-report 45: wrong par $par";
     exit;
 }
 @t=();
 @t=split(/\_/,$par);    # action job
 $action=$t[0];
 $action=~s/[^a-z]//g;

 $id=$t[1];
 $jobid=$id;

 $opt=$ARGV[1];    # p or 123 for debug
 #############################################

 # get par -> %current
 &task_read_jobpar($id);

 #reports to run ex r01, r02   ot R01 on my reports
 $t=$id."_trigger";
 $t=$current{$t};
 $t=~s/ //g;     # remove spaces
 @reports=split(/,/,$t);


 # run?
 if ($action eq "")                     {  print "job-other 42: wrong action - must be not empty\n";      exit; }
 if (!("run,cancel,delete"=~/$action/)) {  print "job-other 43: wrong action - must be run: $action \n";  exit; }

 #cancel or delete
 if ($action eq "cancel") { &my_cancel; $info="cancel job";  &my_log; exit; }
 if ($action eq "delete") { &my_delete; $info="delete job";  exit; }

 if ($action eq "run") {

     $t=&task_jobproc;                  # optionally restart with unique _csperl_id run_id p123
     $t=&task_norestart($id);           # do not run twice, returns ok to continue 

     if ($t ne "ok") {
         print __LINE__." cannot (re)start job: $id $t\n";
         &my_log_id("execution cancelled, job $id is $t");                           # only last.log to preserver error as default
         &my_log_last("execution cancelled, job $id is $t");
         exit;

     } else {

         print "start report $id\n";
         &my_log_id("start report $id");
     }
}

 # read reports with path
    &my_reports;

 #debug
     #&print_array(@reports);

 # run
    $error=&my_run;        # alerts and status
    &my_log;
    exit;
########################### end of main

##############
sub my_reports {
############## active reports with path
    #  read common scripts
    @t=();
    my @r=();
    my $r;

    &read_dir('/xampp/web-gui/data/menues/_lib/scripts/reports/','t','\.pl$');

    #add reports
    foreach my $t (@t) {
      $t="/xampp/web-gui/data/menues/_lib/scripts/reports/$t";
    }
    push (@r,@t);

    # read my scripts
    @t=();
    &read_dir('/xampp/web-gui/_my/scripts/reports/','t','\.pl$');

    #add reports
    foreach my $t (@t) {
      $t="/xampp/web-gui/_my/scripts/reports/$t";
    }
    push (@r,@t);

    ## push active reports to @run
    foreach my $r (@r) {                    # all available reports with path
          foreach my $report (@reports) {   # selected in job ex r01
             if ($r=~/\/$report\#/) {
                push (@run,$r);             # all reports to process
             }
          }
    }
    @run=sort(@run);

    #debug
    # &print_array(@r);   exit;
    # &print_hash(%cfg);

    # same alerts are send only once per day

}

 ## delete job
 sub my_delete {
 ##############
 my ($r);
      if ($id ne "") {
        my @files = glob("/var/web-gui/_log/jobs/$id.*");
        foreach my $f (@files) {
           unlink ($f);
           print "##### job-72: delete $id ######  $r\n";
        }
      } else {
        print "job-74 delete, no id !! ########### \n";
      }
 }

############ update jobpar
sub my_last {
###########
  # update .par
  &task_gettime;                                             # $current{'timestamp'}, $current{'logtime'}
  $t="$id"."_last";                                          # update last
  $current{$t}=$current{'logtime'};
  &task_update_par($id);
}


#############
 sub my_run {
 ############
  my ($r,$todo,$run);
  &getdate;                                #  adminlib

  #debug
  # &print_hash(%current);

  $t=$id."_text1";
  $todo=$current{$t};                       # alert to/ status to

  #run all reports
  my $host=`hostname`;
  chomp($host);


  foreach $run (@run) {

     $r="";

     # for member os, care in reports?

     ##r04: smart optional with now when status is processed
     if (($todo=~/status/i) && ($run=~/smart/i)) {
         print "perl $run now\n";
         $r=`/xampp/perl/bin/perl $run now`;
     } else {
         print "perl $run\n";
         $r=`/xampp/perl/bin/perl $run`;
     }

     if ($r=~/skip: /s)           {  next; }        # skip report
     if (!($r=~/message_begin/s)) {  next; }        # skip report

     # use nl
     $r=~s/<br>/\n/gs;
     print "$r\n\n";

     #process alert and is this an alert
     if ($todo =~/alert/i) {

         #alert allowed?
         if (!($run=~/A/)) {
                print "## no Alert in $run";
                next;
         }

         if (!($r=~/message_begin=ok/s)) {

              #alertlist (multiple error reasons ex jobs)
              if ($r=~/alert_begin=/s) {
                 $t=$r;
                 $t=~s/.*alert_begin=//s;
                 $t=~s/alert_end.*//s;
                 $alertmess.="$t";
              }

              # alerts
              if ($alertmess ne "") {
                   $alert= "Alert/ Error on $host from $sys{'date_and_time'} ($alertmess)\n";
              }

              # get reason
              $t=$r;
              $t=~s/.*message_begin=//s;
              $t=~s/message_end.*//s;

              if ($reasons eq "") {
                 $reasons="Storage alert: ";
              }
              $reasons.="$t, ";            # add current alert reason
              $alert.="\n";               # Alertmessage header

              #get alert message
              $t=$r;
              $t=~s/.*alert_begin=//s;
              $t=~s/alert_end.*//s;
              $alert.="$t\n\n";
         }

     }

     #process status?
     if ($todo =~/status/i) {

         #status allowed?
         if (!($run=~/S/)) {
                print "## no Status in $run";
                next;
         }

         if ($r=~/status_begin=/s) {

              $t=$run; $t=~s/.*\///;
              $text.="\n#######\nReport: $t \n#######\n";         # add current report

              #get status message
              $t=$r;
              $t=~s/.*status_begin=//s;
              $t=~s/status_end.*//s;
              $text.="$t\n\n";
         }
     }
  }

  #debug
  # initiate error (new mail on new error)
  #$alertmess="\ntest2";
  #$alert.="1";

  if ($alertmess eq "") {
    $reasons="none";
  } else {
    $reasons=$alertmess;
  }

  print <<EoF;
########################################################################################
$reports
$text

alerts: $alert\n$reasons

EoF

  $t=$id."_text1";
  if ($current{$t}=~/alert/i) {
       if (!($alert=~/1/)) {
           print "no alerts\n";
           return;
       }
       &my_alert;
  }
  if ($current{$t}=~/status/i) {
       &my_status;
  }


  &my_last;  # update job
  return();
 }


#############
 sub my_alert  {
##############
  # collect alerts

  my ($t,@to,$r,$pools,$w,@l,$l,$tmp,@w,$subject,$sl);

   #sendlog
   my $sl="/xampp/web-gui/_log/reports/a_$jobid\.log";

  $t=$id."_text1";
  $t=$current{$t};                       # status or alert
  if (!($t=~/alert/i)) { return; }       # no alert job

  #send only once a day, rad last errmess with day string within
  # /xampp/web-gui/_log/reports/lastalert.log  # alertmess
  $t=&my_read("$sl");
  $t=~s/.*Alertmessage:\s+//s;   $t=~s/\n+$//s;
  $alertmess=~s/^\n+//s;         $alertmess=~s/\n+$//s;

  print "alertmess=-$alertmess-\nlastalert=-$t-\n";
  if ($t eq $alertmess)  { print "error already send today (err=lasterr)\n"; return; }

  $t=$id."_to";
  my $to=$current{$t};

  #debug
  #&print_hash(%current);

  if (!($alert=~/1/))   { return; }        # no alert
  if ($alertmess eq "") { return; }        # no alert

  my $host=`hostname`;
  chomp($host);

  $alert="Job $id\nAlertmessage: \n$alertmess\n";
  $subject = "*** Storage Alert *** on $host\: ";                # Subject

  #write log
  open (PF, ">$sl");          # ex ex /tmp/nappit/job-scrub_1350378156.request
       flock(PF, LOCK_EX);
       print PF "run_$id reported $sys{'date_us'}\n$subject\n$alert";
  close (PF);

  # if not jets end today
  &my_prepare_send($to,$subject,$alert); # par subject, message

 }

 #############
 sub my_status  {
##############
    # collect status reports

    #sendlog
    my $sl="/xampp/web-gui/_log/reports/s_$jobid\.log";

    $t=$id."_text1";
    $t=$current{$t};
    if (!($t=~/status/i)) { return; }       # no status

    my $host=`hostname`;
    chomp($host);

    $subject="* Storage status report from $host";

    $t=$id."_trigger";
    my $reports=$current{$t};


    my $out=<<EoF;

Statusreport job $id from $host from $sys{'date'}  $sys{'hour'}:$sys{'min'}

################
Storage reports: $reports
################

$text
EoF

$t=$id."_to";
my $to=$current{$t};

#write log
  open (PF, ">$sl");          # ex ex /tmp/nappit/job-scrub_1350378156.request
       flock(PF, LOCK_EX);
       print PF "run_$id $sys{'date'} $sys{'hour'}:$sys{'min'}\n$out\n$text";
  close (PF);

&my_prepare_send($to,$subject,$out); # par subject, message


#    # log status
#    if (-f "/var/web-gui/_log/tmp/mail-status.log") {
#         $text.="\n\n-user defined/ other job status:\n /var/web-gui/_log/tmp/mail-status.log\n\n";
#         $text.=`cat /var/web-gui/_log/tmp/mail-status.log`;
#         if (-f "/var/web-gui/_log/tmp/mail-status.last") { `rm /var/web-gui/_log/tmp/mail-status.last`; }
#         rename("/var/web-gui/_log/tmp/mail-status.log","/var/web-gui/_log/tmp/mail-status.last");
#    }

}  # end my_status


 ###########
 sub my_prepare_send {
 ############

  my ($t,$tls,$host,$send,$r);

  my $to=$_[0];                # allow commalist
  my $subject=$_[1];           # parameter as this is different from alert to status
  my $text   =$_[2];

  print "send job-report $subject\n";

  $t=$id."_send";
  $send=$current{$t};         # tls or basic

  $t=$id."_server";
  my $server=$current{$t};

  $t=$id."_from";
  my $from=$current{$t};

  $t=$id."_user";
  my $user=$current{$t};

  $t=$id."_pw";
  my $pw=$current{$t};

  #webapi
  if ($send=~/webapi/i) {
       # alert only!
       if ($subject=~/status report/) {
         $t=`hostname`; chomp($t);
         $subject="Report from $t";
         $text="use mail for status\nuse webapi for alert only";
       }


       $t=$id."_pushover-token";
       $current{'pushover-token'}=$current{$t};

       $t=$id."_pushover-user";
       $current{'pushover-user'}=$current{$t};

       $current{'subject'}=$subject;
       $current{'message'}=$text;

  }  else {

     #debug
     #  print "server,from,to,sub\n$server, $from, $to\n$subject\n ";

     if (!($server=~/\w+/))    { &my_err('please set a smtp-server in job - setup'); }
     if (!($to=~/\@/))         { &my_err('please set a mailto in job - setup'); }
     if (!($from=~/\@/))       { &my_err('please set a mailfrom in job - setup'); }

  }

  if ($send=~/webapi/i) {

     if (-f "/var/web-gui/_my/scripts/webapi/webapi.pl") {
        $current{'subject'}="my $current{'subject'}";
        require("/var/web-gui/_my/scripts/webapi/webapi.pl");                          # %current: token,to,subject,host,message
     } else {
        require("/var/web-gui/data/napp-it/zfsos/_lib/scripts/webapi.pl");             # %current: token,to,subject,host,message
     }
     print "492 webapi done: $current{'return'}\n";
     $info=$current{'return'};
     &my_log;
     exit;

  } else {
        #tls?
        if ($send=~/tls/i) {
            require("/xampp/web-gui/data/menues/_lib/scripts/reports/smtp-tls.lib");
        } else {
            require("/xampp/web-gui/data/menues/_lib/scripts/reports/smtp-p25.lib");
        }



        $t=$id."_to";
        my $to=$current{$t};                     # opt commalist a@xx.com,b@xx.com,c@xx.com
        my @to=split(/,/,$to);
        foreach my $to (@to) {
             print "## send job email 517 to $to\n";
             $r.=&my_send($server,$from,$to,$subject,$text,$user,$pw);        #server,from,to,sub,text,user,pw
        }
  }
  print "message was send .. \n$r";
  $info=$r;
  &my_log;

 }

############
sub my_err {
############
  $error="error";
  $info=$_[0];

  print "job-report my_err $info\n";

  &my_log;
  exit;
}

###########  ## log last run in jobfile /_log/jobs/id...job
sub my_log {
###########

  # print "job-repli 548 - call log id=$id action=$action\n";

  # line 1: lastrun = ok or err
  # line 2: time of last run
  # line 3+: info

  &task_gettime;                                             # $current{'timestamp'}, $current{'logtime'}

  my (@v,$v,$n,$l,$f1,$f2,@data);

  $l=$error;
  if ($l eq "") { $l="ok"; } else { $l="error"; $info="$info $error"; }

  $l=~s/\n/ /gs;
  $info=~s/\n.*//s;                                        # first line
  $info=~s/ +/ /g;

 #stage 1: log general jobs
 ###########################
  my $log="/xampp/web-gui/_log/jobs/jobs.log";
  my $lock="/xampp/web-gui/_log/jobs/jobs.lock";
  open (X,"<$lock");
  flock(X, LOCK_EX);
          open (F,"<$log");
              flock(F, LOCK_EX);
              @data=<F>;
          close (F);

        #unshift   $l=ok/error
        unshift(@data,"$l; report $ref $id; time: $current{'timestamp'}; info: $info\n");

        if (@data > 50) {
                   $t=@data-50;
                   splice(@data,50,$t);
        }

       $t=join("",@data);
       open (PF, ">$log");
          flock(PF, LOCK_EX);
          print PF "$t";
       close (PF);
 close (X);

 if ($info=~/^delete /) { exit; }


 #stage 2 update job.par  $current{'logtime'};
  $log="/xampp/web-gui/_log/jobs/$id.log";

  $t="$id"."_last";                                              # update last
  $current{$t}=$current{'logtime'};
  &task_update_par($id);                                         # update last


  # stage 3 id.log newest at top
  ###################
  &my_log_id("$l","info: $info");

  # update .par
  &task_gettime;                                             # $current{'timestamp'}, $current{'logtime'}
  $t="$id"."_last";                                              # update last
  $current{$t}=$current{'logtime'};
  &task_update_par($id);

 exit;  # log always last action

}

##
1;